"""Defines the basic Equation and all of its support machinery including
the Group class.
"""
# System library imports.
import ast

from collections import defaultdict

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import re
from copy import deepcopy
import inspect
import itertools
import numpy
from textwrap import dedent, wrap

from compyle.api import (CythonGenerator, KnownType,
                         OpenCLConverter, get_symbols)
from compyle.translator import CUDAConverter
from compyle.config import get_config


getfullargspec = getattr(
    inspect, 'getfullargspec', inspect.getargspec
)


def camel_to_underscore(name):
    """Given a CamelCase name convert it to a name with underscores,
    i.e. camel_case.
    """
    # From stackoverflow: :P
    # http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-camel-case
    s1 = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def indent(text, prefix='    '):
    """Prepend prefix to every line in the text"""
    return ''.join(prefix + line for line in text.splitlines(True))


##############################################################################
# `Context` class.
##############################################################################
class Context(dict):
    """Based on the Bunch receipe by Alex Martelli from Active State's recipes.

    A convenience class used to specify a context in which a code block will
    execute.

    Example
    -------

    Basic usage::
        >>> c = Context(a=1, b=2)
        >>> c.a
        1
        >>> c.x = 'a'
        >>> c.x
        'a'
        >>> c.keys()
        ['a', 'x', 'b']
    """

    def __getattr__(self, key):
        try:
            return self.__getitem__(key)
        except KeyError:
            raise AttributeError('Context has no attribute %s' % key)

    def __setattr__(self, key, value):
        self[key] = value


def get_array_names(symbols):
    """Given a set of symbols, return a set of source array names and
    a set of destination array names.
    """
    src_arrays = set(x for x in symbols
                     if x.startswith('s_') and x != 's_idx')
    dest_arrays = set(x for x in symbols
                      if x.startswith('d_') and x != 'd_idx')
    return src_arrays, dest_arrays


##############################################################################
# `BasicCodeBlock` class.
##############################################################################
class BasicCodeBlock(object):
    """Encapsulates a string of code and the context in which it executes.

    It also performs some simple analysis of the code that proves handy.
    """

    ##########################################################################
    # `object` interface.
    ##########################################################################
    def __init__(self, code, **kwargs):
        """Constructor.

        Parameters
        ----------

        code : str: source code.
        kwargs : values which define the context of the code.
        """
        self.setup(code, **kwargs)

    def __call__(self, **kwargs):
        """A simplistic test for the code that runs the code in the setup
        context with any additional arguments passed set in the context.

        Note that this will make a deepcopy of the context to prevent any
        changes to the original context.

        It returns a dictionary.

        """
        context = deepcopy(dict(self.context))
        if kwargs:
            context.update(kwargs)
        bytecode = compile(self.code, '<string>', 'exec')
        glb = globals()
        exec(bytecode, glb, context)
        return Context(**context)

    ##########################################################################
    # Private interface.
    ##########################################################################
    def _setup_context(self):
        context = self.context
        symbols = self.symbols
        for index in ('s_idx', 'd_idx'):
            if index in symbols and index not in context:
                context[index] = 0

        for a_name in itertools.chain(self.src_arrays, self.dest_arrays):
            if a_name not in context:
                context[a_name] = numpy.zeros(2, dtype=float)

    def _setup_code(self, code):
        """Perform analysis of the code and store the information in various
        attributes.
        """
        code = dedent(code)
        self.code = code
        self.ast_tree = ast.parse(code)
        self.symbols = get_symbols(self.ast_tree)

        symbols = self.symbols
        self.src_arrays, self.dest_arrays = get_array_names(symbols)
        self._setup_context()

    ##########################################################################
    # Public interface.
    ##########################################################################
    def setup(self, code, **kwargs):
        """Setup the code and context with the given arguments.

        Parameters
        ----------

        code : str: source code.

        kwargs : values which define the context of the code.
        """
        self.context = Context(**kwargs)

        if code is not None:
            self._setup_code(code)


##############################################################################
# Convenient precomputed symbols and their code.
##############################################################################
def precomputed_symbols():
    """Return a collection of predefined symbols that can be used in equations.
    """
    c = Context()
    c.HIJ = BasicCodeBlock(code="HIJ = 0.5*(d_h[d_idx] + s_h[s_idx])", HIJ=0.0)

    c.EPS = BasicCodeBlock(code="EPS = 0.01*HIJ*HIJ", EPS=0.0)

    c.RHOIJ = BasicCodeBlock(code="RHOIJ = 0.5*(d_rho[d_idx] + s_rho[s_idx])",
                             RHOIJ=0.0)

    c.RHOIJ1 = BasicCodeBlock(code="RHOIJ1 = 1.0/RHOIJ", RHOIJ1=0.0)

    c.XIJ = BasicCodeBlock(
        code=dedent(
            """
            XIJ[0] = d_x[d_idx] - s_x[s_idx]
            XIJ[1] = d_y[d_idx] - s_y[s_idx]
            XIJ[2] = d_z[d_idx] - s_z[s_idx]
            """
        ),
        XIJ=[0.0, 0.0, 0.0]
    )

    c.VIJ = BasicCodeBlock(
        code=dedent(
            """
            VIJ[0] = d_u[d_idx] - s_u[s_idx]
            VIJ[1] = d_v[d_idx] - s_v[s_idx]
            VIJ[2] = d_w[d_idx] - s_w[s_idx]
            """
        ),
        VIJ=[0.0, 0.0, 0.0]
    )

    c.R2IJ = BasicCodeBlock(
        code=dedent(
            """
            R2IJ = XIJ[0]*XIJ[0] + XIJ[1]*XIJ[1] + XIJ[2]*XIJ[2]
            """
        ),
        R2IJ=0.0
    )

    c.RIJ = BasicCodeBlock(code="RIJ = sqrt(R2IJ)", RIJ=0.0)

    c.WIJ = BasicCodeBlock(
        code="WIJ = KERNEL(XIJ, RIJ, HIJ)",
        WIJ=0.0
    )

    # wdeltap for tensile instability correction
    c.WDP = BasicCodeBlock(
        code="WDP = KERNEL(XIJ, DELTAP*HIJ, HIJ)",
        WDP=0.0
    )

    c.WI = BasicCodeBlock(
        code="WI = KERNEL(XIJ, RIJ, d_h[d_idx])",
        WI=0.0
    )

    c.WJ = BasicCodeBlock(
        code="WJ = KERNEL(XIJ, RIJ, s_h[s_idx])",
        WJ=0.0
    )

    c.WDASHI = BasicCodeBlock(
        code="WDASHI = DWDQ(RIJ, d_h[d_idx])",
        WDASHI=0.0
    )

    c.WDASHJ = BasicCodeBlock(
        code="WDASHJ = DWDQ(RIJ, s_h[s_idx])",
        WDASHJ=0.0
    )

    c.WDASHIJ = BasicCodeBlock(
        code="WDASHIJ = DWDQ(RIJ, HIJ)",
        WDASHIJ=0.0
    )

    c.DWIJ = BasicCodeBlock(
        code="GRADIENT(XIJ, RIJ, HIJ, DWIJ)",
        DWIJ=[0.0, 0.0, 0.0]
    )

    c.DWI = BasicCodeBlock(
        code="GRADIENT(XIJ, RIJ, d_h[d_idx], DWI)",
        DWI=[0.0, 0.0, 0.0]
    )

    c.DWJ = BasicCodeBlock(
        code="GRADIENT(XIJ, RIJ, s_h[s_idx], DWJ)",
        DWJ=[0.0, 0.0, 0.0]
    )

    c.GHI = BasicCodeBlock(
        code="GHI = GRADH(XIJ, RIJ, d_h[d_idx])",
        GHI=0.0
    )

    c.GHJ = BasicCodeBlock(
        code="GHJ = GRADH(XIJ, RIJ, s_h[s_idx])",
        GHJ=0.0
    )

    c.GHIJ = BasicCodeBlock(code="GHIJ = GRADH(XIJ, RIJ, HIJ)", GHIJ=0.0)

    return c


def sort_precomputed(precomputed, all_pre_comp):
    """Sorts the precomputed equations in the given dictionary as per the
    dependencies of the symbols and returns an ordered dict.

    Note that this will not deal with finding any precomputed symbols that
    are dependent on other precomputed symbols.  It only sorts them in the
    right order.
    """
    weights = dict((x, None) for x in precomputed)
    pre_comp = all_pre_comp
    # Find the dependent pre-computed symbols for each in the precomputed.
    depends = dict((x, None) for x in precomputed)
    for pre, cb in precomputed.items():
        depends[pre] = [x for x in cb.symbols if x in pre_comp and x != pre]

    # The basic algorithm is to assign weights to each of the precomputed
    # symbols based on the maximum weight of the dependencies of the
    # precomputed symbols.  This way, those with no dependencies have weight
    # zero and those with more have heigher weights. The `levels` dict stores
    # a list of precomputed symbols for each  weight.  These are then stored
    # in an ordered dict in the order of the weights to produce the output.
    levels = defaultdict(list)
    pre_comp_names = list(precomputed.keys())
    while pre_comp_names:
        for name in pre_comp_names[:]:
            wts = [weights[x] for x in depends[name]]
            if len(wts) == 0:
                weights[name] = 0
                levels[0].append(name)
                pre_comp_names.remove(name)
            elif None in wts:
                continue
            else:
                level = max(wts) + 1
                weights[name] = level
                levels[level].append(name)
                pre_comp_names.remove(name)

    result = OrderedDict()
    for level in range(len(levels)):
        for name in sorted(levels[level]):
            result[name] = pre_comp[name]

    return result


def get_predefined_types(precomp):
    """Return a dictionary that can be used by a CythonGenerator for
    the precomputed symbols.
    """
    result = {'dt': 0.0,
              't': 0.0,
              'dst': KnownType('object'),
              'NBRS': KnownType('unsigned int*'),
              'N_NBRS': KnownType('int'),
              'src': KnownType('ParticleArrayWrapper')}
    for sym, value in precomp.items():
        result[sym] = value.context[sym]
    return result


def get_arrays_used_in_equation(equation):
    """Return two sets, the source and destination arrays used by the equation.
    """
    src_arrays = set()
    dest_arrays = set()
    methods = (
        'initialize', 'initialize_pair', 'loop', 'loop_all', 'post_loop'
    )
    for meth_name in methods:
        meth = getattr(equation, meth_name, None)
        if meth is not None:
            args = getfullargspec(meth).args
            s, d = get_array_names(args)
            src_arrays.update(s)
            dest_arrays.update(d)
    return src_arrays, dest_arrays


def get_init_args(obj, method, ignore=None):
    """Return the arguments for the method given, typically an __init__.
    """
    ignore = ignore if ignore is not None else []
    spec = getfullargspec(method)
    keys = [k for k in spec.args[1:] if k not in ignore and k in obj.__dict__]
    args = ['%s=%r' % (k, getattr(obj, k)) for k in keys]
    return args


##############################################################################
# `Equation` class.
##############################################################################
class Equation(object):
    ##########################################################################
    # `object` interface.
    ##########################################################################
    def __init__(self, dest, sources):
        r"""
        Parameters
        ----------
        dest : str
            name of the destination particle array
        sources : list of str or None
            names of the source particle arrays
        """
        self.dest = dest
        if sources is not None and len(sources) > 0:
            self.sources = sources
        else:
            self.sources = None

        # Does the equation require neighbors or not.
        self.no_source = self.sources is None
        self.name = self.__class__.__name__
        # The name of the variable used in the compiled AccelerationEval
        # instance.
        self.var_name = ''

    def __repr__(self):
        name = self.__class__.__name__
        args = get_init_args(self, self.__init__, [])
        res = '%s(%s)' % (name, ', '.join(args))
        return '\n'.join(wrap(res, width=70, break_long_words=False))

    def converged(self):
        """Return > 0 to indicate converged iterations and < 0 otherwise.
        """
        return 1.0

    def _pull(self, *args):
        """Pull attributes from the GPU if needed.

        The GPU reduce and converged methods run on the host and not on
        the device and this is useful to call there.  This is not useful
        on the CPU as this does not matter which is why this is a
        private method.
        """
        if hasattr(self, '_gpu'):
            ary = self._gpu.get()
            if len(args) == 0:
                args = ary.dtype.names
            for arg in args:
                setattr(self, arg, ary[arg][0])


###############################################################################
# `Group` class.
###############################################################################
class Group(object):
    """A group of equations.

    This class provides some support for the code generation for the
    collection of equations.
    """

    pre_comp = precomputed_symbols()

    def __init__(self, equations, real=True, update_nnps=False, iterate=False,
                 max_iterations=1, min_iterations=0, pre=None, post=None,
                 condition=None, start_idx=0, stop_idx=None):
        """Constructor.

        Parameters
        ----------

        equations: list
            a list of equation objects.

        real: bool
            specifies if only non-remote/non-ghost particles should be
            operated on.

        update_nnps: bool
            specifies if the neighbors should be re-computed locally after
            this group

        iterate: bool
            specifies if the group should continue iterating until each
            equation's "converged()" methods returns with a positive value.

        max_iterations: int
            specifies the maximum number of times this group should be
            iterated.

        min_iterations: int
            specifies the minimum number of times this group should be
            iterated.

        pre: callable
            A callable which is passed no arguments that is called before
            anything in the group is executed.

        post: callable
            A callable which is passed no arguments that is called after
            the group is completed.

        condition: callable
            A callable that is passed (t, dt). If this callable returns True,
            the group is executed, otherwise it is not. If condition is None,
            the group is always executed. Note that this should work even if
            the group has many destination arrays.

        start_idx: int or str
            Start looping from this destination index. Starts from the given
            number if an integer is passed. If a string is look for a
            property/constant and use its first value as the loop count.

        stop_idx: int or str
            Loop up to this destination index instead of over all possible
            values. Defaults to all particles. Ends at the given number if an
            integer is passed. If a string is passed, look for a
            property/constant and use its first value as the loop count. Note
            that this works like a range stop parameter so the last value is
            not included.

        Notes
        -----

        When running simulations in parallel, one should typically
        run the summation density over all particles (both local and remote)
        in each processor.  This is because we must update the
        pressure/density of the remote neighbors in the current processor.
        Otherwise the results can be incorrect with the remote particles
        having an older density.  This is also the case for the TaitEOS.  In
        these cases the group that computes the equation should set real to
        False.

        """
        self.real = real
        self.update_nnps = update_nnps

        # iterative groups
        self.iterate = iterate
        self.max_iterations = max_iterations
        self.min_iterations = min_iterations
        self.pre = pre
        self.post = post
        self.condition = condition
        self.start_idx = start_idx
        self.stop_idx = stop_idx

        only_groups = [x for x in equations if isinstance(x, Group)]
        if (len(only_groups) > 0) and (len(only_groups) != len(equations)):
            raise ValueError(
                'All elements must be Groups if you use sub groups.'
            )

        # This group has only sub-groups.
        self.has_subgroups = len(only_groups) > 0

        self.equations = equations
        self.src_arrays = self.dest_arrays = None

        self.update()

    ##########################################################################
    # Non-public interface.
    ##########################################################################
    def __repr__(self):
        cls = self.__class__.__name__
        eqs = ', \n'.join(repr(eq) for eq in self.equations)
        ignore = ['equations']
        if self.start_idx != 0:
            ignore.append('start_idx')
        for prop in ['pre', 'post', 'condition', 'stop_idx']:
            if getattr(self, prop) is None:
                ignore.append(prop)
        kws = ', '.join(get_init_args(self, self.__init__, ignore))
        kws = '\n'.join(wrap(kws, width=74, subsequent_indent=' '*2,
                             break_long_words=False))
        return '%s(equations=[\n%s\n  ],\n  %s)' % (
            cls, indent(eqs), kws
        )

    def _has_code(self, kind='loop'):
        assert kind in ('initialize', 'initialize_pair', 'loop', 'loop_all',
                        'post_loop', 'reduce')
        for equation in self.equations:
            if hasattr(equation, kind):
                return True

    def _setup_precomputed(self):
        """Get the precomputed symbols for this group of equations.
        """
        # Calculate the precomputed symbols for this equation.
        all_args = set()
        for equation in self.equations:
            if hasattr(equation, 'loop'):
                args = getfullargspec(equation.loop).args
                all_args.update(args)
        all_args.discard('self')

        pre = self.pre_comp
        precomputed = dict((s, pre[s]) for s in all_args if s in pre)

        # Now find the precomputed symbols in the pre-computed symbols.
        done = False
        found_precomp = set(precomputed.keys())
        while not done:
            done = True
            all_new = set()
            for sym in found_precomp:
                code_block = pre[sym]
                new = set([s for s in code_block.symbols
                           if s in pre and s not in precomputed])
                all_new.update(new)
            if len(all_new) > 0:
                done = False
                for s in all_new:
                    precomputed[s] = pre[s]
            found_precomp = all_new

        self.precomputed = sort_precomputed(precomputed, pre)

        # Update the context.
        context = self.context
        for p, cb in self.precomputed.items():
            context[p] = cb.context[p]

    ##########################################################################
    # Public interface.
    ##########################################################################
    def update(self):
        self.context = Context()
        if not self.has_subgroups:
            self._setup_precomputed()

    def get_array_names(self, recompute=False):
        """Returns two sets of array names, the first being source_arrays
        and the second being destination array names.
        """
        if not recompute and self.src_arrays is not None:
            return set(self.src_arrays), set(self.dest_arrays)
        src_arrays = set()
        dest_arrays = set()
        for equation in self.equations:
            s, d = get_arrays_used_in_equation(equation)
            src_arrays.update(s)
            dest_arrays.update(d)

        for cb in self.precomputed.values():
            src_arrays.update(cb.src_arrays)
            dest_arrays.update(cb.dest_arrays)

        self.src_arrays = src_arrays
        self.dest_arrays = dest_arrays
        return src_arrays, dest_arrays

    def get_converged_condition(self):
        if self.has_subgroups:
            code = [g.get_converged_condition() for g in self.equations]
            return ' & '.join(code)
        else:
            code = []
            for equation in self.equations:
                code.append('(self.%s.converged() > 0)' % equation.var_name)
            # Note, we use '&' because we want to call converged on all
            # equations and not be short-circuited by the first one that
            # returns False.
            return ' & '.join(code)

    def get_variable_names(self):
        # First get all the contexts and find the names.
        all_vars = set()
        for cb in self.precomputed.values():
            all_vars.update(cb.symbols)

        # Filter out all arrays.
        filtered_vars = [x for x in all_vars
                         if not x.startswith(('s_', 'd_'))]
        # Filter other things.
        ignore = ['KERNEL', 'GRADIENT', 's_idx', 'd_idx']
        # Math functions.
        import math
        ignore += [x for x in dir(math) if not x.startswith('_')
                   and callable(getattr(math, x))]
        try:
            ignore.remove('gamma')
            ignore.remove('lgamma')
        except ValueError:
            # Older Python's don't have gamma/lgamma.
            pass
        filtered_vars = [x for x in filtered_vars if x not in ignore]

        return filtered_vars

    def has_initialize(self):
        return self._has_code('initialize')

    def has_initialize_pair(self):
        return self._has_code('initialize_pair')

    def has_loop(self):
        return self._has_code('loop')

    def has_loop_all(self):
        return self._has_code('loop_all')

    def has_post_loop(self):
        return self._has_code('post_loop')

    def has_reduce(self):
        return self._has_code('reduce')


class CythonGroup(Group):
    ##########################################################################
    # Non-public interface.
    ##########################################################################
    def _get_variable_decl(self, context, mode='declare'):
        decl = []
        names = list(context.keys())
        names.sort()
        for var in names:
            value = context[var]
            if isinstance(value, int):
                declare = 'cdef long ' if mode == 'declare' else ''
                decl.append('{declare}{var} = {value}'.format(declare=declare,
                                                              var=var,
                                                              value=value))
            elif isinstance(value, float):
                declare = 'cdef double ' if mode == 'declare' else ''
                decl.append('{declare}{var} = {value}'.format(declare=declare,
                                                              var=var,
                                                              value=value))
            elif isinstance(value, (list, tuple)):
                if mode == 'declare':
                    decl.append(
                        'cdef DoubleArray _{var} = '
                        'DoubleArray(aligned({size}, 8)*self.n_threads)'
                        .format(
                            var=var, size=len(value)
                        )
                    )
                    decl.append('cdef double* {var} = _{var}.data'
                                .format(size=len(value), var=var))
                else:
                    pass
        return '\n'.join(decl)

    def _get_code(self, kernel=None, kind='loop'):
        assert kind in ('initialize', 'initialize_pair', 'loop', 'loop_all',
                        'post_loop', 'reduce')
        # We assume here that precomputed quantities are only relevant
        # for loops and not post_loops and initialization.
        pre = []
        if kind == 'loop':
            for p, cb in self.precomputed.items():
                pre.append(cb.code.strip())
            if len(pre) > 0:
                pre.extend(['', ''])
        preamble = self._set_kernel('\n'.join(pre), kernel)

        code = []
        for eq in self.equations:
            meth = getattr(eq, kind, None)
            if meth is not None:
                args = getfullargspec(meth).args
                if 'self' in args:
                    args.remove('self')
                if 'SPH_KERNEL' in args:
                    args[args.index('SPH_KERNEL')] = 'self.kernel'
                if kind == 'reduce':
                    args = ['dst.array', 't', 'dt']
                call_args = ', '.join(args)
                c = 'self.{eq_name}.{method}({args})' \
                    .format(eq_name=eq.var_name, method=kind, args=call_args)
                code.append(c)
        if len(code) > 0:
            code.append('')
        return preamble + '\n'.join(code)

    def _set_kernel(self, code, kernel):
        if kernel is not None:
            k_func = 'self.kernel.kernel'
            w_func = 'self.kernel.dwdq'
            g_func = 'self.kernel.gradient'
            h_func = 'self.kernel.gradient_h'
            deltap = 'self.kernel.get_deltap()'
            code = code.replace('DELTAP', deltap)
            return code.replace('GRADIENT', g_func).replace(
                'KERNEL', k_func
            ).replace('GRADH', h_func).replace('DWDQ', w_func)
        else:
            return code

    ##########################################################################
    # Public interface.
    ##########################################################################
    def get_array_declarations(self, names, known_types={}):
        decl = []
        for arr in sorted(names):
            if arr in known_types:
                decl.append('cdef {type} {arr}'.format(
                    type=known_types[arr].type, arr=arr
                ))
            else:
                decl.append('cdef double* %s' % arr)
        return '\n'.join(decl)

    def get_variable_declarations(self, context):
        return self._get_variable_decl(context, mode='declare')

    def get_variable_array_setup(self):
        names = list(self.context.keys())
        names.sort()
        code = []
        for var in names:
            value = self.context[var]
            if isinstance(value, (list, tuple)):
                code.append(
                    '{var} = &_{var}.data[thread_id*aligned({size}, 8)]'
                    .format(size=len(value), var=var)
                )
        return '\n'.join(code)

    def get_initialize_code(self, kernel=None):
        return self._get_code(kernel, kind='initialize')

    def get_initialize_pair_code(self, kernel=None):
        return self._get_code(kernel, kind='initialize_pair')

    def get_loop_code(self, kernel=None):
        return self._get_code(kernel, kind='loop')

    def get_loop_all_code(self, kernel=None):
        return self._get_code(kernel, kind='loop_all')

    def get_post_loop_code(self, kernel=None):
        return self._get_code(kernel, kind='post_loop')

    def get_py_initialize_code(self):
        lines = []
        for i, equation in enumerate(self.equations):
            if hasattr(equation, 'py_initialize'):
                code = ('self.all_equations["{name}"].py_initialize'
                        '(dst.array, t, dt)').format(name=equation.var_name)
                lines.append(code)
        return '\n'.join(lines)

    def get_reduce_code(self):
        return self._get_code(kernel=None, kind='reduce')

    def get_equation_wrappers(self, known_types={}):
        classes = defaultdict(lambda: 0)
        eqs = {}
        for equation in self.equations:
            cls = equation.__class__.__name__
            n = classes[cls]
            equation.var_name = '%s%d' % (
                camel_to_underscore(equation.name), n
            )
            classes[cls] += 1
            eqs[cls] = equation
        wrappers = []
        predefined = dict(get_predefined_types(self.pre_comp))
        predefined.update(known_types)
        code_gen = CythonGenerator(known_types=predefined)
        for cls in sorted(classes.keys()):
            code_gen.parse(eqs[cls])
            wrappers.append(code_gen.get_code())
        return '\n'.join(wrappers)

    def get_equation_defs(self):
        lines = []
        for equation in self.equations:
            code = 'cdef public {cls} {name}'.format(cls=equation.name,
                                                     name=equation.var_name)
            lines.append(code)
        return '\n'.join(lines)

    def get_equation_init(self):
        lines = []
        for i, equation in enumerate(self.equations):
            code = 'self.{name} = {cls}(**equations[{idx}].__dict__)' \
                .format(name=equation.var_name, cls=equation.name,
                        idx=i)
            lines.append(code)
        return '\n'.join(lines)


class OpenCLGroup(Group):
    _Converter_Class = OpenCLConverter

    # #### Private interface  #####
    def _update_for_local_memory(self, predefined, eqs):
        modified_classes = []
        loop_ann = predefined.copy()
        for k in loop_ann.keys():
            if 's_' in k:
                # TODO: Make each argument have their own KnownType
                # right from the start
                new_type = loop_ann[k].type.replace(
                    'GLOBAL_MEM', 'LOCAL_MEM'
                ).replace('__global', 'LOCAL_MEM')
                loop_ann[k] = KnownType(new_type)
        for eq in eqs.values():
            cls = eq.__class__
            loop = getattr(cls, 'loop', None)
            if loop is not None:
                self._set_loop_annotation(loop, loop_ann)
                modified_classes.append(cls)
        return modified_classes

    def _set_loop_annotation(self, func, value):
        try:
            func.__annotations__ = value
        except AttributeError:
            func.im_func.__annotations__ = value

    ##########################################################################
    # Public interface.
    ##########################################################################
    def get_equation_wrappers(self, known_types={}):
        classes = defaultdict(lambda: 0)
        eqs = {}
        for equation in self.equations:
            cls = equation.__class__.__name__
            n = classes[cls]
            equation.var_name = '%s%d' % (
                camel_to_underscore(equation.name), n
            )
            classes[cls] += 1
            eqs[cls] = equation
        wrappers = []
        predefined = dict(get_predefined_types(self.pre_comp))
        predefined.update(known_types)
        predefined['NBRS'] = KnownType('GLOBAL_MEM unsigned int*')

        use_local_memory = get_config().use_local_memory
        modified_classes = []
        if use_local_memory:
            modified_classes = self._update_for_local_memory(predefined, eqs)

        code_gen = self._Converter_Class(known_types=predefined)
        ignore = ['reduce', 'converged']
        for cls in sorted(classes.keys()):
            src = code_gen.parse_instance(eqs[cls], ignore_methods=ignore)
            wrappers.append(src)

        if use_local_memory:
            # Remove the added annotations
            for cls in modified_classes:
                self._set_loop_annotation(cls.loop, {})

        return '\n'.join(wrappers)


class CUDAGroup(OpenCLGroup):
    _Converter_Class = CUDAConverter


class MultiStageEquations(object):
    '''A class that allows a user to specify different equations
    for different stages.

    The object doesn't do much, except contain the different collections of
    equations.

    '''

    def __init__(self, groups):
        '''
        Parameters
        ----------

        groups: list/tuple
            A list/tuple of list of groups/equations, one for each stage.

        '''
        assert type(groups) in (list, tuple)
        self.groups = groups

    def __repr__(self):
        name = self.__class__.__name__
        groups = [', \n'.join(str(stg_grps) for stg_grps in stg)
                  for stg in self.groups]
        kw = ""
        for i, group in enumerate(groups):
            stage = i
            kw += '[\n# Stage %d\n' % stage
            kw += group
            kw += '\n# End Stage %d\n],\n' % stage

        s = '%s(groups=[\n%s])' % (
            name, indent(kw, '  '),
        )
        return s

    def __len__(self):
        return len(self.groups)



# cython: language_level=3, language=c++, embedsignature=True
# Library imports.
import numpy as np
cimport numpy as np

# Cython imports
from cython.operator cimport dereference as deref, preincrement as inc
from cython.parallel import parallel, prange, threadid

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.vector cimport vector

# cpython
from cpython.dict cimport PyDict_Clear, PyDict_Contains, PyDict_GetItem
from cpython.list cimport PyList_GetItem, PyList_SetItem, PyList_GET_ITEM

# Cython for compiler directives
cimport cython


#############################################################################
cdef class BoxSortNNPS(LinkedListNNPS):

    """Nearest neighbor query class using the box sort method but which
    uses the LinkedList algorithm.  This makes this very fast but
    perhaps not as safe as the DictBoxSortNNPS.  All this class does
    is to use a std::map to obtain a linear cell index from the actual
    flattened cell index.
    """

    #### Private protocol ################################################

    cdef long _get_flattened_cell_index(self, cPoint pnt, double cell_size):
        cdef long cell_id = flatten(
            find_cell_id(pnt, cell_size), self.ncells_per_dim, self.dim
        )
        return self.cell_to_index[cell_id]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef inline long _get_valid_cell_index(self, int cid_x, int cid_y, int cid_z,
            int* ncells_per_dim, int dim, int n_cells) nogil:
        """Return the flattened index for a valid cell"""
        cdef long ncx = ncells_per_dim[0]
        cdef long ncy = ncells_per_dim[1]
        cdef long ncz = ncells_per_dim[2]

        cdef long cell_index = -1

        # basic test for valid indices. Since we bin the particles with
        # respect to the origin, negative indices can never occur.
        cdef bint is_valid = (
            (ncx > cid_x > -1) and (ncy > cid_y > -1) and (ncz > cid_z > -1)
        )

        # Given the validity of the cells, return the flattened cell index
        cdef map[long, int].iterator it
        if is_valid:
            cell_id = flatten_raw(cid_x, cid_y, cid_z, ncells_per_dim, dim)
            if cell_id > -1:
                it = self.cell_to_index.find(cell_id)
                if it != self.cell_to_index.end():
                    cell_index = deref(it).second

        return cell_index

    cpdef long _count_occupied_cells(self, long n_cells) except -1:
        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int np
        cdef DoubleArray x, y, z
        cdef DoubleArray xmin = self.xmin
        cdef DoubleArray xmax = self.xmax
        cdef IntArray ncells_per_dim = self.ncells_per_dim
        cdef int narrays = self.narrays
        cdef int dim = self.dim
        cdef double cell_size = self.cell_size

        cdef cPoint pnt = cPoint_new(0, 0, 0)
        cdef long _cid
        cdef map[long, int] _cid_to_index
        cdef pair[long, int] _entry

        # flattened cell index
        cdef int i, j
        cdef long cell_index

        for j in range(narrays):
            pa_wrapper = pa_wrappers[j]
            np = pa_wrapper.get_number_of_particles()
            x = pa_wrapper.x
            y = pa_wrapper.y
            z = pa_wrapper.z

            for i in range(np):
                # the flattened index is considered relative to the
                # minimum along each co-ordinate direction
                pnt.x = x.data[i] - xmin.data[0]
                pnt.y = y.data[i] - xmin.data[1]
                pnt.z = z.data[i] - xmin.data[2]

                # flattened cell index
                _cid = flatten( find_cell_id( pnt, cell_size ), ncells_per_dim, dim )
                _entry.first = _cid
                _entry.second = 1
                _cid_to_index.insert(_entry)

        cdef map[long, int].iterator it = _cid_to_index.begin()
        cdef int count = 0
        while it != _cid_to_index.end():
            _entry = deref(it)
            _cid_to_index[_entry.first] = count
            count += 1
            inc(it)

        self.cell_to_index = _cid_to_index
        return count

##############################################################################
cdef class DictBoxSortNNPS(NNPS):
    """Nearest neighbor query class using the box-sort algorithm using a
    dictionary.

    NNPS bins all local particles using the box sort algorithm in
    Cells. The cells are stored in a dictionary 'cells' which is keyed
    on the spatial index (IntPoint) of the cell.

    """
    def __init__(self, int dim, list particles, double radius_scale=2.0,
                 int ghost_layers=1, domain=None, cache=False,
                 sort_gids=False):
        """Constructor for NNPS

        Parameters
        ----------

        dim : int
            Number of dimensions.

        particles : list
            The list of particle arrays we are working on.

        radius_scale : double, default (2)
            Optional kernel radius scale. Defaults to 2

        domain : DomainManager, default (None)
            Optional limits for the domain

        cache : bint
            Flag to set if we want to cache neighbor calls. This costs
            storage but speeds up neighbor calculations.

        sort_gids : bint, default (False)
            Flag to sort neighbors using gids (if they are available).
            This is useful when comparing parallel results with those
            from a serial run.
        """
        # initialize the base class
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        # initialize the cells dict
        self.cells = {}

        # compute the intial box sort. First, the Domain Manager's
        # update method is called to comptue the maximum smoothing
        # length for particle binning.
        self.domain.update()
        self.update()

        msg = """WARNING: The cache will currently work only if
        find_nearest_neighbors works and can be used without requiring
        the GIL.  The DictBoxSort does not work with OpenMP since it uses
        Python objects and cannot release the GIL.  Until this is fixed,
        we cannot use caching, or parallel neighbor finding using this
        DictBoxSortNNPS, use the more efficient LinkedListNNPS instead.
        Disabling caching for now.
        """
        if cache:
            print(msg)
        self.use_cache = False


    #### Public protocol ################################################

    cpdef get_nearest_particles_no_cache(self, int src_index, int dst_index,
                            size_t d_idx, UIntArray nbrs, bint prealloc):
        """Utility function to get near-neighbors for a particle.

        Parameters
        ----------

        src_index : int
            Index of the source particle array in the particles list

        dst_index : int
            Index of the destination particle array in the particles list

        d_idx : int (input)
            Destination particle for which neighbors are sought.

        nbrs : UIntArray (output)
            Neighbors for the requested particle are stored here.

        prealloc : bool
            Specifies if the neighbor array already has pre-allocated space
            for the neighbor list.  In this case the neighbors are directly set
            in the given array without resetting or appending to the array.
            This improves performance when the neighbors are cached.
        """
        cdef dict cells = self.cells
        cdef Cell cell

        cdef NNPSParticleArrayWrapper src = self.pa_wrappers[ src_index ]
        cdef NNPSParticleArrayWrapper dst = self.pa_wrappers[ dst_index ]

        # Source data arrays
        cdef DoubleArray s_x = src.x
        cdef DoubleArray s_y = src.y
        cdef DoubleArray s_z = src.z
        cdef DoubleArray s_h = src.h
        cdef UIntArray s_gid = src.gid

        # Destination particle arrays
        cdef DoubleArray d_x = dst.x
        cdef DoubleArray d_y = dst.y
        cdef DoubleArray d_z = dst.z
        cdef DoubleArray d_h = dst.h
        cdef UIntArray d_gid = dst.gid

        cdef double radius_scale = self.radius_scale
        cdef double cell_size = self.cell_size
        cdef UIntArray lindices
        cdef size_t indexj, count
        cdef ZOLTAN_ID_TYPE j

        cdef cPoint xi = cPoint_new(d_x.data[d_idx], d_y.data[d_idx], d_z.data[d_idx])
        cdef cIntPoint _cid = find_cell_id( xi, cell_size )
        cdef IntPoint cid = IntPoint_from_cIntPoint( _cid )
        cdef IntPoint cellid = IntPoint(0, 0, 0)

        cdef double xij2

        cdef double hi2, hj2

        hi2 = radius_scale * d_h.data[d_idx]
        hi2 *= hi2

        cdef int ierr

        # reset the nbr array length. This should avoid a realloc
        if prealloc:
            nbrs.length = 0
        else:
            nbrs.reset()
        count = 0

        cdef int ix, iy, iz
        for ix in [cid.data.x -1, cid.data.x, cid.data.x + 1]:
            for iy in [cid.data.y - 1, cid.data.y, cid.data.y + 1]:
                for iz in [cid.data.z - 1, cid.data.z, cid.data.z + 1]:
                    cellid.data.x = ix; cellid.data.y = iy; cellid.data.z = iz

                    ierr = PyDict_Contains( cells, cellid )
                    if ierr == 1:

                        cell = <Cell>PyDict_GetItem( cells, cellid )
                        lindices = <UIntArray>PyList_GetItem( cell.lindices, src_index )

                        for indexj in range( lindices.length ):
                            j = lindices.data[indexj]

                            xij2 = norm2( s_x.data[j]-xi.x,
                                          s_y.data[j]-xi.y,
                                          s_z.data[j]-xi.z )

                            hj2 = radius_scale * s_h.data[j]
                            hj2 *= hj2

                            # select neighbor
                            if ( (xij2 < hi2) or (xij2 < hj2) ):
                                if prealloc:
                                    nbrs.data[count] = j
                                    count += 1
                                else:
                                    nbrs.append( j )
        if prealloc:
            nbrs.length = count

        if self.sort_gids:
            self._sort_neighbors(nbrs.data, count, s_gid.data)

    #### Private protocol ################################################

    cpdef _refresh(self):
        """Clear the cells dict"""
        cdef dict cells = self.cells
        PyDict_Clear( cells )

    cpdef _bin(self, int pa_index, UIntArray indices):
        """Bin a given particle array with indices.

        Parameters
        ----------

        pa_index : int
            Index of the particle array corresponding to the particles list

        indices : UIntArray
            Subset of particles to bin

        """
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[ pa_index ]
        cdef DoubleArray x = pa_wrapper.x
        cdef DoubleArray y = pa_wrapper.y
        cdef DoubleArray z = pa_wrapper.z

        cdef dict cells = self.cells
        cdef double cell_size = self.cell_size

        cdef UIntArray lindices, gindices
        cdef size_t num_particles, indexi, i

        cdef cIntPoint _cid
        cdef IntPoint cid

        cdef Cell cell
        cdef int ierr, narrays = self.narrays

        # now begin binning the particles
        num_particles = indices.length
        for indexi in range(num_particles):
            i = indices.data[indexi]

            pnt = cPoint_new( x.data[i], y.data[i], z.data[i] )
            _cid = find_cell_id( pnt, cell_size )

            cid = IntPoint_from_cIntPoint(_cid)

            ierr = PyDict_Contains(cells, cid)
            if ierr == 0:
                cell = Cell(cid, cell_size, narrays)
                cells[ cid ] = cell

            # add this particle to the list of indicies
            cell = <Cell>PyDict_GetItem( cells, cid )
            lindices = <UIntArray>PyList_GetItem( cell.lindices, pa_index )

            lindices.append( <ZOLTAN_ID_TYPE> i )
            #gindices.append( gid.data[i] )

        self.n_cells = <int>len(cells)
        self._cell_keys = list(cells.keys())



"""
Flow past cylinder
"""
import numpy as np
import os

from pysph.base.kernels import QuinticSpline
from pysph.sph.equation import Equation
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.tools import geometry as G
from pysph.sph.wc.edac import EDACScheme
from pysph.sph.bc.inlet_outlet_manager import (
        InletInfo, OutletInfo)

# Fluid mechanical/numerical parameters
rho = 1000
umax = 1.0
c0 = 10 * umax
p0 = rho * c0 * c0


class SolidWallNoSlipBCReverse(Equation):
    def __init__(self, dest, sources, nu):
        self.nu = nu
        super(SolidWallNoSlipBCReverse, self).__init__(dest, sources)

    def initialize(self, d_idx, d_auf, d_avf, d_awf):
        d_auf[d_idx] = 0.0
        d_avf[d_idx] = 0.0
        d_awf[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, s_rho, d_V, s_V,
             d_ug, d_vg, d_wg,
             d_auf, d_avf, d_awf,
             s_u, s_v, s_w,
             DWIJ, R2IJ, EPS, XIJ):

        # averaged shear viscosity Eq. (6).
        etai = self.nu * d_rho[d_idx]
        etaj = self.nu * s_rho[s_idx]

        etaij = 2 * (etai * etaj)/(etai + etaj)

        # particle volumes; d_V inverse volume.
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # scalar part of the kernel gradient
        Fij = XIJ[0]*DWIJ[0] + XIJ[1]*DWIJ[1] + XIJ[2]*DWIJ[2]

        # viscous contribution (third term) from Eq. (8), with VIJ
        # defined appropriately using the ghost values
        tmp = 1./d_m[d_idx] * (Vi2 + Vj2) * (etaij * Fij/(R2IJ + EPS))

        d_auf[d_idx] += tmp * (d_ug[d_idx] - s_u[s_idx])
        d_avf[d_idx] += tmp * (d_vg[d_idx] - s_v[s_idx])
        d_awf[d_idx] += tmp * (d_wg[d_idx] - s_w[s_idx])


class ResetInletVelocity(Equation):
    def __init__(self, dest, sources, U, V, W):
        self.U = U
        self.V = V
        self.W = W

        super(ResetInletVelocity, self).__init__(dest, sources)

    def loop(self, d_idx, d_u, d_v, d_w, d_x, d_y, d_z, d_xn, d_yn, d_zn,
             d_uref):
        if d_idx == 0:
            d_uref[0] = self.U
        d_u[d_idx] = self.U
        d_v[d_idx] = self.V
        d_w[d_idx] = self.W


class WindTunnel(Application):
    def initialize(self):
        # Geometric parameters
        self.Lt = 30.0  # length of tunnel
        self.Wt = 15.0  # half width of tunnel
        self.dc = 1.2  # diameter of cylinder
        self.cxy = 10., 0.0  # center of cylinder
        self.nl = 10  # Number of layers for wall/inlet/outlet
        self.io_method = 'donothing'

    def add_user_options(self, group):
        group.add_argument(
            "--re", action="store", type=float, dest="re", default=200,
            help="Reynolds number."
        )
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.2,
            help="Ratio h/dx."
        )
        group.add_argument(
            "--nx", action="store", type=int, dest="nx", default=12,
            help="Number of points in 1D of the cylinder."
        )
        group.add_argument(
            "--lt", action="store", type=float, dest="Lt", default=30,
            help="Length of the WindTunnel."
        )
        group.add_argument(
            "--wt", action="store", type=float, dest="Wt", default=15,
            help="Half width of the WindTunnel."
        )
        group.add_argument(
            "--dc", action="store", type=float, dest="dc", default=1.2,
            help="Diameter of the cylinder."
        )
        group.add_argument(
            "--io-method", action="store", type=str, dest="io_method",
            default='donothing', help="'donothing', 'mirror',"
            "or 'characteristic', 'mod_donothing', hybrid."
        )

    def consume_user_options(self):
        self.dc = dc = self.options.dc
        self.Lt = self.options.Lt/2 * dc
        self.Wt = self.options.Wt/2 * dc
        self.io_method = self.options.io_method
        nx = self.options.nx
        re = self.options.re

        self.nu = nu = umax * self.dc / re

        self.cxy = 5.*self.dc, 0
        self.dx = dx = self.dc / nx
        self.volume = dx * dx
        hdx = self.options.hdx
        self.nl = (int)(6.0*hdx)

        self.h = h = hdx * self.dx
        dt_cfl = 0.25 * h / (c0 + umax)
        dt_viscous = 0.125 * h**2 / nu

        self.dt = min(dt_cfl, dt_viscous)
        self.tf = 100.0

    def _create_fluid(self):
        dx = self.dx
        h0 = self.h
        x, y = np.mgrid[dx / 2:self.Lt:dx, -self.Wt + dx/2:self.Wt:dx]
        x, y = (np.ravel(t) for t in (x, y))
        one = np.ones_like(x)
        volume = dx * dx * one
        m = volume * rho
        fluid = get_particle_array(
            name='fluid', m=m, x=x, y=y, h=h0, V=1.0 / volume, u=umax,
            p=0.0, rho=rho)
        return fluid

    def _create_solid(self):
        dx = self.dx
        h0 = self.h
        x = [0.0]
        y = [0.0]
        r = dx
        nt = 0
        while r - self.dc / 2 < 0.00001:
            nnew = int(np.pi*r**2/dx**2 + 0.5)
            tomake = nnew-nt
            theta = np.linspace(0., 2.*np.pi, tomake + 1)
            for t in theta[:-1]:
                x.append(r*np.cos(t))
                y.append(r*np.sin(t))
            nt = nnew
            r = r + dx
        x = np.array(x)
        y = np.array(y)
        x, y = (t.ravel() for t in (x, y))
        x += self.cxy[0]
        volume = dx*dx
        solid = get_particle_array(
            name='solid', x=x, y=y,
            m=volume*rho, rho=rho, h=h0, V=1.0/volume)
        return solid

    def _create_wall(self):
        dx = self.dx
        h0 = self.h
        x0, y0 = np.mgrid[
            dx/2: self.Lt+self.nl*dx+self.nl*dx: dx, dx/2: self.nl*dx: dx]
        x0 -= self.nl*dx
        y0 -= self.nl*dx+self.Wt
        x0 = np.ravel(x0)
        y0 = np.ravel(y0)

        x1 = np.copy(x0)
        y1 = np.copy(y0)
        y1 += self.nl*dx+2*self.Wt
        x1 = np.ravel(x1)
        y1 = np.ravel(y1)

        x0 = np.concatenate((x0, x1))
        y0 = np.concatenate((y0, y1))
        volume = dx*dx
        wall = get_particle_array(
            name='wall', x=x0, y=y0, m=volume*rho, rho=rho, h=h0,
            V=1.0/volume)
        return wall

    def _set_wall_normal(self, pa):
        props = ['xn', 'yn', 'zn']
        for p in props:
            pa.add_property(p)

        y = pa.y
        cond = y > 0.0
        pa.yn[cond] = 1.0
        cond = y < 0.0
        pa.yn[cond] = -1.0

    def _create_outlet(self):
        dx = self.dx
        h0 = self.h
        x, y = np.mgrid[dx/2:self.nl * dx:dx,  -self.Wt + dx/2:self.Wt:dx]
        x, y = (np.ravel(t) for t in (x, y))
        x += self.Lt
        one = np.ones_like(x)
        volume = dx * dx * one
        m = volume * rho
        outlet = get_particle_array(
            name='outlet', x=x, y=y, m=m, h=h0, V=1.0/volume, u=umax,
            p=0.0, rho=one * rho, uhat=umax)
        return outlet

    def _create_inlet(self):
        dx = self.dx
        h0 = self.h
        x, y = np.mgrid[dx / 2:self.nl*dx:dx, -self.Wt + dx/2:self.Wt:dx]
        x, y = (np.ravel(t) for t in (x, y))
        x = x - self.nl * dx
        one = np.ones_like(x)
        volume = one * dx * dx

        inlet = get_particle_array(
            name='inlet', x=x, y=y, m=volume * rho, h=h0, u=umax, rho=rho,
            V=1.0 / volume, p=0.0)
        return inlet

    def create_particles(self):
        dx = self.dx
        fluid = self._create_fluid()
        solid = self._create_solid()
        G.remove_overlap_particles(fluid, solid, dx, dim=2)
        outlet = self._create_outlet()
        inlet = self._create_inlet()
        wall = self._create_wall()

        ghost_inlet = self.iom.create_ghost(inlet, inlet=True)
        ghost_outlet = self.iom.create_ghost(outlet, inlet=False)

        particles = [fluid, inlet, outlet, solid, wall]
        if ghost_inlet:
            particles.append(ghost_inlet)
        if ghost_outlet:
            particles.append(ghost_outlet)

        self.scheme.setup_properties(particles)
        self._set_wall_normal(wall)

        if self.io_method == 'hybrid':
            fluid.uag[:] = umax
            fluid.uta[:] = umax
            outlet.uta[:] = umax

        return particles

    def create_scheme(self):
        h = nu = None
        s = EDACScheme(
            ['fluid'], ['solid'], dim=2, rho0=rho, c0=c0, h=h, pb=p0,
            nu=nu, inlet_outlet_manager=None,
            inviscid_solids=['wall']
        )
        return s

    def configure_scheme(self):
        scheme = self.scheme
        self.iom = self._create_inlet_outlet_manager()
        scheme.inlet_outlet_manager = self.iom
        pfreq = 100
        kernel = QuinticSpline(dim=2)
        self.iom.update_dx(self.dx)
        scheme.configure(h=self.h, nu=self.nu)

        scheme.configure_solver(kernel=kernel, tf=self.tf, dt=self.dt,
                                pfreq=pfreq, n_damp=0)

    def _get_io_info(self):
        inleteqns = [ResetInletVelocity('ghost_inlet', [], U=-umax, V=0.0,
                     W=0.0),
                     ResetInletVelocity('inlet', [], U=umax, V=0.0,
                     W=0.0)]

        i_update_cls = None
        i_has_ghost = True
        o_update_cls = None
        o_has_ghost = True
        manager = None
        props_to_copy = ['x0', 'y0', 'z0', 'uhat', 'vhat', 'what', 'x', 'y',
                         'z', 'u', 'v', 'w', 'm', 'h', 'rho', 'p', 'ioid']

        if self.io_method == 'donothing':
            from pysph.sph.bc.donothing.inlet import Inlet
            from pysph.sph.bc.donothing.outlet import Outlet
            from pysph.sph.bc.donothing.simple_inlet_outlet import (
                SimpleInletOutlet)
            o_has_ghost = False
            i_update_cls = Inlet
            o_update_cls = Outlet
            manager = SimpleInletOutlet
        elif self.io_method == 'mirror':
            from pysph.sph.bc.mirror.inlet import Inlet
            from pysph.sph.bc.mirror.outlet import Outlet
            from pysph.sph.bc.mirror.simple_inlet_outlet import (
                SimpleInletOutlet)
            i_update_cls = Inlet
            o_update_cls = Outlet
            manager = SimpleInletOutlet
        elif self.io_method == 'hybrid':
            from pysph.sph.bc.hybrid.inlet import Inlet
            from pysph.sph.bc.hybrid.outlet import Outlet
            from pysph.sph.bc.hybrid.simple_inlet_outlet import (
                SimpleInletOutlet)
            i_update_cls = Inlet
            o_update_cls = Outlet
            o_has_ghost = False
            manager = SimpleInletOutlet
            props_to_copy += ['uta', 'pta', 'u0', 'v0', 'w0', 'p0']
        if self.io_method == 'mod_donothing':
            from pysph.sph.bc.mod_donothing.inlet import Inlet
            from pysph.sph.bc.mod_donothing.outlet import Outlet
            from pysph.sph.bc.mod_donothing.simple_inlet_outlet import (
                SimpleInletOutlet)
            o_has_ghost = False
            i_update_cls = Inlet
            o_update_cls = Outlet
            manager = SimpleInletOutlet
        if self.io_method == 'characteristic':
            from pysph.sph.bc.characteristic.inlet import Inlet
            from pysph.sph.bc.characteristic.outlet import Outlet
            from pysph.sph.bc.characteristic.simple_inlet_outlet import (
                SimpleInletOutlet)
            o_has_ghost = False
            i_update_cls = Inlet
            o_update_cls = Outlet
            manager = SimpleInletOutlet

        inlet_info = InletInfo(
            pa_name='inlet', normal=[-1.0, 0.0, 0.0],
            refpoint=[0.0, 0.0, 0.0], equations=inleteqns,
            has_ghost=i_has_ghost, update_cls=i_update_cls,
            umax=umax
            )

        outlet_info = OutletInfo(
            pa_name='outlet', normal=[1.0, 0.0, 0.0],
            refpoint=[self.Lt, 0.0, 0.0], has_ghost=o_has_ghost,
            update_cls=o_update_cls, equations=None,
            props_to_copy=props_to_copy
        )

        return inlet_info, outlet_info, manager

    def _create_inlet_outlet_manager(self):
        inlet_info, outlet_info, manager = self._get_io_info()
        iom = manager(
            fluid_arrays=['fluid'], inletinfo=[inlet_info],
            outletinfo=[outlet_info]
        )
        return iom

    def create_inlet_outlet(self, particle_arrays):
        iom = self.iom
        io = iom.get_inlet_outlet(particle_arrays)
        return io

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return
        t, cd, cl = self._plot_force_vs_t()
        res = os.path.join(self.output_dir, 'results.npz')
        np.savez(res, t=t, cd=cd, cl=cl)

    def _plot_force_vs_t(self):
        from pysph.solver.utils import iter_output, load
        from pysph.tools.sph_evaluator import SPHEvaluator
        from pysph.sph.equation import Group
        from pysph.base.kernels import QuinticSpline
        from pysph.sph.wc.transport_velocity import (
            MomentumEquationPressureGradient,
            SummationDensity, SetWallVelocity
        )

        data = load(self.output_files[0])
        solid = data['arrays']['solid']
        fluid = data['arrays']['fluid']

        prop = ['awhat', 'auhat', 'avhat', 'wg', 'vg', 'ug', 'V', 'uf', 'vf',
                'wf', 'wij', 'vmag', 'pavg', 'nnbr', 'auf', 'avf', 'awf']
        for p in prop:
            solid.add_property(p)
            fluid.add_property(p)

        # We find the force of the solid on the fluid and the opposite of that
        # is the force on the solid. Note that the assumption is that the solid
        # is far from the inlet and outlet so those are ignored.
        print(self.nu, p0, self.dc, rho)
        equations = [
            Group(
                equations=[
                    SummationDensity(dest='fluid', sources=['fluid', 'solid']),
                    SummationDensity(dest='solid', sources=['fluid', 'solid']),
                    SetWallVelocity(dest='solid', sources=['fluid']),
                    ], real=False),
            Group(
                equations=[
                    # Pressure gradient terms
                    MomentumEquationPressureGradient(
                        dest='solid', sources=['fluid'], pb=p0),
                    SolidWallNoSlipBCReverse(
                        dest='solid', sources=['fluid'], nu=self.nu),
                    ], real=True),
        ]
        sph_eval = SPHEvaluator(
            arrays=[solid, fluid], equations=equations, dim=2,
            kernel=QuinticSpline(dim=2)
        )
        t, cd, cl = [], [], []
        import gc
        print(self.dc, self.dx, self.nu)
        print('fxf', 'fxp', 'fyf', 'fyp', 'cd', 'cl', 't')
        for sd, arrays in iter_output(self.output_files[:]):
            fluid = arrays['fluid']
            solid = arrays['solid']
            for p in prop:
                solid.add_property(p)
                fluid.add_property(p)
            t.append(sd['t'])
            sph_eval.update_particle_arrays([solid, fluid])
            sph_eval.evaluate()
            fxp = sum(solid.m*solid.au)
            fyp = sum(solid.m*solid.av)
            fxf = sum(solid.m*solid.auf)
            fyf = sum(solid.m*solid.avf)
            fx = fxf + fxp
            fy = fyf + fyp
            cd.append(fx/(0.5 * rho * umax**2 * self.dc))
            cl.append(fy/(0.5 * rho * umax**2 * self.dc))
            print(fxf, fxp, fyf, fyp, cd[-1], cl[-1], t[-1])
            gc.collect()
        t, cd, cl = list(map(np.asarray, (t, cd, cl)))
        # Now plot the results.
        import matplotlib
        matplotlib.use('Agg')
        from matplotlib import pyplot as plt
        plt.figure()
        plt.plot(t, cd, label=r'$C_d$')
        plt.plot(t, cl, label=r'$C_l$')
        plt.xlabel(r'$t$')
        plt.ylabel('cd/cl')
        plt.legend()
        plt.grid()
        fig = os.path.join(self.output_dir, "force_vs_t.png")
        plt.savefig(fig, dpi=300)
        plt.close()
        return t, cd, cl

    def customize_output(self):
        if self.io_method == 'hybrid':
            self._mayavi_config('''
            viewer.scalar = 'u'
            ''')
        elif self.io_method == 'mirror':
            self._mayavi_config('''
            viewer.scalar = 'u'
            parr = ['ghost_outlet', 'ghost_inlet']
            for particle in parr:
                b = particle_arrays[particle]
                b.visible = False
            ''')
        else:
            self._mayavi_config('''
            viewer.scalar = 'u'
            parr = ['ghost_inlet']
            for particle in parr:
                b = particle_arrays[particle]
                b.visible = False
            ''')

    def post_step(self, solver):
        freq = 500
        if solver.count % freq == 0:
            self.nnps.update()
            for i, pa in enumerate(self.particles):
                if pa.name == 'fluid':
                    self.nnps.spatially_order_particles(i)
            self.nnps.update()


if __name__ == '__main__':
    app = WindTunnel()
    app.run()
    app.post_process(app.info_filename)



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from libcpp.map cimport map

from .nnps_base cimport *
from .linked_list_nnps cimport *

# NNPS using the original gridding algorithm
cdef class DictBoxSortNNPS(NNPS):
    cdef public dict cells               # lookup table for the cells
    cdef list _cell_keys

    cpdef get_nearest_particles_no_cache(self, int src_index, int dst_index,
            size_t d_idx, UIntArray nbrs, bint prealloc)

    cpdef _refresh(self)

    cpdef _bin(self, int pa_index, UIntArray indices)

# NNPS using the linked list approach
cdef class BoxSortNNPS(LinkedListNNPS):
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef public map[long, int] cell_to_index  # Maps cell ID to an index



import numpy as np
import os

from pysph.sph.surface_tension import get_surface_tension_equations
from pysph.tools.geometry import get_2d_block
from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline, QuinticSpline

from pysph.solver.application import Application
from pysph.solver.solver import Solver

from pysph.sph.integrator_step import TransportVelocityStep
from pysph.sph.integrator import PECIntegrator

from pysph.base.nnps import DomainManager
from pysph.solver.utils import iter_output


dim = 2
Lx = 0.5
Ly = 1.0
factor1 = 0.8
factor2 = 1.0/factor1
nu = 0.0
sigma = 1.0
rho0 = 1.

c0 = 20.0
gamma = 1.4
R = 287.1

p0 = c0**2 * rho0

nx = 50
dx = Lx / nx
volume = dx * dx
hdx = 1.5

h0 = hdx * dx

tf = 0.5

epsilon = 0.01 / h0

KE = 10**(-6.6)*p0*p0*gamma/(c0 * c0 * rho0 * rho0 * nx * nx * (gamma - 1))

Vmax = np.sqrt(2 * KE / (rho0 * dx * dx))

dt1 = 0.25*np.sqrt(rho0*h0*h0*h0/(2.0*np.pi*sigma))

dt2 = 0.25*h0/(c0+Vmax)

dt = 0.9*min(dt1, dt2)


class MultiPhase(Application):
    def add_user_options(self, group):
        choices = ['morris', 'tvf', 'adami_stress', 'adami', 'shadloo']
        group.add_argument(
            "--scheme", action="store", dest='scheme', default='morris',
            choices=choices,
            help='Specify scheme to use among %s' % choices
        )

    def create_particles(self):
        fluid_x, fluid_y = get_2d_block(
            dx=dx, length=Lx-dx, height=Ly-dx, center=np.array([0., 0.5*Ly]))
        rho_fluid = np.ones_like(fluid_x) * rho0
        m_fluid = rho_fluid * volume
        h_fluid = np.ones_like(fluid_x) * h0
        cs_fluid = np.ones_like(fluid_x) * c0
        additional_props = ['V', 'color', 'scolor', 'cx', 'cy', 'cz',
                            'cx2', 'cy2', 'cz2', 'nx', 'ny', 'nz', 'ddelta',
                            'uhat', 'vhat', 'what', 'auhat', 'avhat', 'awhat',
                            'ax', 'ay', 'az', 'wij', 'vmag2', 'N', 'wij_sum',
                            'rho0', 'u0', 'v0', 'w0', 'x0', 'y0', 'z0',
                            'kappa', 'arho', 'nu', 'pi00', 'pi01', 'pi02',
                            'pi10', 'pi11', 'pi12', 'pi20', 'pi21', 'pi22']
        fluid = get_particle_array(
            name='fluid', x=fluid_x, y=fluid_y, h=h_fluid, m=m_fluid,
            rho=rho_fluid, cs=cs_fluid, additional_props=additional_props)
        for i in range(len(fluid.x)):
            if fluid.y[i] > 0.25 and fluid.y[i] < 0.75:
                fluid.color[i] = 1.0
            else:
                fluid.color[i] = 0.0
        fluid.V[:] = 1. / volume
        fluid.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                 'ddelta', 'kappa', 'N', 'scolor', 'p'])
        angles = np.random.random_sample((len(fluid.x),))*2*np.pi
        vel = np.sqrt(2 * KE / fluid.m)
        fluid.u = vel
        fluid.v = vel
        fluid.nu[:] = 0.0
        return [fluid]

    def create_domain(self):
        return DomainManager(
            xmin=-0.5 * Lx, xmax=0.5 * Lx, ymin=0.0, ymax=Ly,
            periodic_in_x=True, periodic_in_y=True, n_layers=6)

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = PECIntegrator(fluid=TransportVelocityStep())
        solver = Solver(
            kernel=kernel, dim=dim, integrator=integrator,
            dt=dt, tf=tf, adaptive_timestep=False)
        return solver

    def create_equations(self):
        return get_surface_tension_equations(['fluid'], [],
                                             self.options.scheme, rho0, p0, c0,
                                             0, factor1, factor2, nu, sigma, 2,
                                             epsilon, gamma, real=False)

    def post_process(self):
        try:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
        except ImportError:
            print("Post processing requires Matplotlib")
            return
        from pysph.solver.utils import load
        files = self.output_files
        ke = []
        t = []
        for f in files:
            data = load(f)
            pa = data['arrays']['fluid']
            t.append(data['solver_data']['t'])
            m = pa.m
            u = pa.u
            v = pa.v
            length = len(m)
            ke.append(np.log10(sum(0.5 * m * (u**2 + v**2) / length)))
        fname = os.path.join(self.output_dir, 'results.npz')
        np.savez(fname, t=t, ke=ke)
        plt.plot(t, ke, 'o')
        fig = os.path.join(self.output_dir, "KEvst.png")
        plt.savefig(fig)
        plt.close()


if __name__ == '__main__':
    app = MultiPhase()
    app.run()
    app.post_process()



""" Helper functions to generate commonly used geometries.

PySPH used an axis convention as follows:

Y
|
|
|
|
|
|      /Z
|     /
|    /
|   /
|  /
| /
|/_________________X



"""

import numpy

def create_2D_tank(x1,y1,x2,y2,dx):
    """ Generate an open rectangular tank.

    Parameters:
    -----------

    x1,y1,x2,y2 : Coordinates defining the rectangle in 2D

    dx : The spacing to use

    """

    yl = numpy.arange(y1, y2+dx/2, dx)
    xl = numpy.ones_like(yl) * x1
    nl = len(xl)

    yr = numpy.arange(y1,y2+dx/2, dx)
    xr = numpy.ones_like(yr) * x2
    nr = len(xr)

    xb = numpy.arange(x1+dx, x2-dx+dx/2, dx)
    yb = numpy.ones_like(xb) * y1
    nb = len(xb)

    n = nb + nl + nr

    x = numpy.empty( shape=(n,) )
    y = numpy.empty( shape=(n,) )

    idx = 0
    x[idx:nl] = xl; y[idx:nl] = yl

    idx += nl
    x[idx:idx+nb] = xb; y[idx:idx+nb] = yb

    idx += nb
    x[idx:idx+nr] = xr; y[idx:idx+nr] = yr

    return x, y

def create_3D_tank(x1, y1, z1, x2, y2, z2, dx):
    """ Generate an open rectangular tank.

    Parameters:
    -----------

    x1,y1,x2,y2,x3,y3 : Coordinates defining the rectangle in 2D

    dx : The spacing to use

    """

    points = []
    # create the base X-Y plane
    x, y = numpy.mgrid[x1:x2+dx/2:dx, y1:y2+dx/2:dx]
    x = x.ravel(); y = y.ravel()
    z = numpy.ones_like(x) * z1

    for i in range(len(x)):
        points.append( (x[i], y[i], z[i]) )

    # create the front X-Z plane
    x, z = numpy.mgrid[x1:x2+dx/2:dx, z1:z2+dx/2:dx]
    x = x.ravel(); z = z.ravel()
    y = numpy.ones_like(x) * y1

    for i in range(len(x)):
        points.append( (x[i], y[i], z[i]) )

    # create the Y-Z plane
    y, z = numpy.mgrid[y1:y2+dx/2:dx, z1:z2+dx/2:dx]
    y = y.ravel(); z = z.ravel()
    x = numpy.ones_like(y) * x1

    for i in range(len(x)):
        points.append( (x[i], y[i], z[i]) )

    # create the second X-Z plane
    x, z = numpy.mgrid[x1:x2+dx/2:dx, z1:z2+dx/2:dx]
    x = x.ravel(); z = z.ravel()
    y = numpy.ones_like(x) * y2

    for i in range(len(x)):
        points.append( (x[i], y[i], z[i]) )

    # create the second Y-Z plane
    y, z = numpy.mgrid[y1:y2+dx/2:dx, z1:z2+dx/2:dx]
    y = y.ravel(); z = z.ravel()
    x = numpy.ones_like(y) * x2

    for i in range(len(x)):
        points.append( (x[i], y[i], z[i]) )

    points = set(points)

    x = numpy.array( [i[0] for i in points] )
    y = numpy.array( [i[1] for i in points] )
    z = numpy.array( [i[2] for i in points] )

    return x, y, z

def create_2D_filled_region(x1, y1, x2, y2, dx):
    x,y = numpy.mgrid[x1:x2+dx/2:dx, y1:y2+dx/2:dx]
    x = x.ravel(); y = y.ravel()

    return x, y

def create_3D_filled_region(x1, y1, z1, x2, y2, z2, dx):
    x,y,z = numpy.mgrid[x1:x2+dx/2:dx, y1:y2+dx/2:dx, z1:z2+dx/2:dx]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()

    return x, y, z





"""unittests for the serial NNPS

You can run the tests like so:

    $ pytest -v test_nnps.py
"""
import numpy
from numpy import random

# PySPH imports
from pysph.base.point import IntPoint, Point
from pysph.base.utils import get_particle_array
from pysph.base import nnps
from compyle.config import get_config

# Carrays from PyZoltan
from cyarray.carray import UIntArray, IntArray

# Python testing framework
import unittest
import pytest
from pytest import importorskip, mark
from pysph.base.nnps import get_number_of_threads


class SimpleNNPSTestCase(unittest.TestCase):
    """Simplified NNPS test case

    We distribute particles manually and perform sanity checks on NNPS

    """

    def setUp(self):
        """Default set-up used by all the tests

        Particles with the following coordinates (x, y, z) are placed in a box

        0 : -1.5 , 0.25 , 0.5
        1 : 0.33 , -0.25, 0.25
        2 : 1.25 , -1.25, 1.25
        3 : 0.05 , 1.25 , -0.5
        4 : -0.5 , 0.5  , -1.25
        5 : -0.75, 0.75 , -1.25
        6 : -1.25, 0.5  , 0.5
        7 : 0.5  , 1.5  , -0.5
        8 : 0.5  , -0.5 , 0.5
        9 : 0.5  , 1.75 , -0.75

        The cell size is set to 1. Valid cell indices and the
        particles they contain are given below:

        (-2, 0, 0) : particle 0, 6
        (0, -1, 0) : particle 1, 8
        (1, -2, 1) : particle 2
        (0, 1, -1) : particle 3, 7, 9
        (-1, 0, -2): particle 4, 5

        """
        x = numpy.array([
            -1.5, 0.33, 1.25, 0.05, -0.5, -0.75, -1.25, 0.5, 0.5, 0.5])

        y = numpy.array([
            0.25, -0.25, -1.25, 1.25, 0.5, 0.75, 0.5, 1.5, -0.5, 1.75])

        z = numpy.array([
            0.5, 0.25, 1.25, -0.5, -1.25, -1.25, 0.5, -0.5, 0.5, -0.75])

        # using a degenrate (h=0) array will set cell size to 1 for NNPS
        h = numpy.zeros_like(x)

        pa = get_particle_array(x=x, y=y, z=z, h=h)

        self.dict_box_sort_nnps = nnps.DictBoxSortNNPS(
            dim=3, particles=[pa], radius_scale=1.0
        )

        self.box_sort_nnps = nnps.BoxSortNNPS(
            dim=3, particles=[pa], radius_scale=1.0
        )

        self.ll_nnps = nnps.LinkedListNNPS(
            dim=3, particles=[pa], radius_scale=1.0
        )

        self.sp_hash_nnps = nnps.SpatialHashNNPS(
            dim=3, particles=[pa], radius_scale=1.0
        )

        self.ext_sp_hash_nnps = nnps.ExtendedSpatialHashNNPS(
            dim=3, particles=[pa], radius_scale=1.0
        )

        self.strat_radius_nnps = nnps.StratifiedHashNNPS(
            dim=3, particles=[pa], radius_scale=1.0
        )

        # these are the expected cells
        self.expected_cells = {
            IntPoint(-2, 0, 0): [0, 6],
            IntPoint(0, -1, 0): [1, 8],
            IntPoint(1, -2, 1): [2],
            IntPoint(0, 1, -1): [3, 7, 9],
            IntPoint(-1, 0, -2): [4, 5]
        }

    def test_cell_size(self):
        "SimpleNNPS :: test cell_size"
        nnps = self.dict_box_sort_nnps
        self.assertAlmostEqual(nnps.cell_size, 1.0, 14)

        nnps = self.box_sort_nnps
        self.assertAlmostEqual(nnps.cell_size, 1.0, 14)

        nnps = self.ll_nnps
        self.assertAlmostEqual(nnps.cell_size, 1.0, 14)

        nnps = self.sp_hash_nnps
        self.assertAlmostEqual(nnps.cell_size, 1.0, 14)

        nnps = self.ext_sp_hash_nnps
        self.assertAlmostEqual(nnps.cell_size, 1.0, 14)

        nnps = self.strat_radius_nnps
        self.assertAlmostEqual(nnps.cell_size, 1.0, 14)

    def test_cells(self):
        "SimpleNNPS :: test cells"
        nnps = self.dict_box_sort_nnps
        cells = self.expected_cells

        # check each cell for it's contents
        for key in cells:
            self.assertTrue(key in nnps.cells)

            cell = nnps.cells.get(key)

            cell_indices = list(cell.lindices[0].get_npy_array())
            expected_indices = cells.get(key)

            self.assertTrue(cell_indices == expected_indices)


class NNPS2DTestCase(unittest.TestCase):
    def setUp(self):
        """Default set-up used by all the tests

        Two sets of particle arrays (a & b) are created and neighbors
        are checked from a -> b, b -> a , a -> a and b -> b

        """
        numpy.random.seed(123)
        self.numPoints1 = numPoints1 = 1 << 11
        self.numPoints2 = numPoints2 = 1 << 10

        self.pa1 = pa1 = self._create_random(numPoints1)
        self.pa2 = pa2 = self._create_random(numPoints2)

        # the list of particles
        self.particles = [pa1, pa2]

    def _create_random(self, numPoints):
        # average particle spacing and volume in the unit cube
        dx = pow(1.0 / numPoints, 1. / 2.)

        # create random points in the interval [-1, 1]^3
        x1, y1 = random.random((2, numPoints)) * 2.0 - 1.0
        z1 = numpy.zeros_like(x1)

        h1 = numpy.ones_like(x1) * 1.2 * dx
        gid1 = numpy.arange(numPoints).astype(numpy.uint32)

        # first particle array
        pa = get_particle_array(
            x=x1, y=y1, z=z1, h=h1, gid=gid1)

        return pa

    def _assert_neighbors(self, nbrs_nnps, nbrs_brute_force):
        # ensure that the lengths of the arrays are the same
        self.assertEqual(nbrs_nnps.length, nbrs_brute_force.length)
        nnbrs = nbrs_nnps.length

        _nbrs1 = nbrs_nnps.get_npy_array()
        _nbrs2 = nbrs_brute_force.get_npy_array()

        # sort the neighbors
        nbrs1 = sorted(_nbrs1[:nnbrs])
        nbrs2 = _nbrs2
        nbrs2.sort()

        # check each neighbor
        for i in range(nnbrs):
            self.assertEqual(nbrs1[i], nbrs2[i])

    def _test_neighbors_by_particle(self, src_index, dst_index, dst_numPoints):
        # nnps and the two neighbor lists
        nps = self.nps
        nbrs1 = UIntArray()
        nbrs2 = UIntArray()

        nps.set_context(src_index, dst_index)

        # get the neighbors and sort the result
        for i in range(dst_numPoints):
            nps.get_nearest_particles(src_index, dst_index, i, nbrs1)
            nps.brute_force_neighbors(src_index, dst_index, i, nbrs2)

            # ensure that the neighbor lists are the same
            self._assert_neighbors(nbrs1, nbrs2)


class DictBoxSortNNPS2DTestCase(NNPS2DTestCase):
    """Test for the original box-sort algorithm"""

    def setUp(self):
        NNPS2DTestCase.setUp(self)
        self.nps = nnps.DictBoxSortNNPS(
            dim=2, particles=self.particles, radius_scale=2.0
        )

    def test_neighbors_aa(self):
        self._test_neighbors_by_particle(src_index=0, dst_index=0,
                                         dst_numPoints=self.numPoints1)

    def test_neighbors_ab(self):
        self._test_neighbors_by_particle(src_index=0, dst_index=1,
                                         dst_numPoints=self.numPoints2)

    def test_neighbors_ba(self):
        self._test_neighbors_by_particle(src_index=1, dst_index=0,
                                         dst_numPoints=self.numPoints1)

    def test_neighbors_bb(self):
        self._test_neighbors_by_particle(src_index=1, dst_index=1,
                                         dst_numPoints=self.numPoints2)

    def test_repeated(self):
        self.test_neighbors_aa()
        self.test_neighbors_ab()
        self.test_neighbors_ba()
        self.test_neighbors_bb()


class OctreeGPUNNPS2DTestCase(DictBoxSortNNPS2DTestCase):
    """Test for Z-Order SFC based OpenCL algorithm"""

    def setUp(self):
        NNPS2DTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps

        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False

        self.nps = gpu_nnps.OctreeGPUNNPS(
            dim=2, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )

    def tearDown(self):
        super(OctreeGPUNNPS2DTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class OctreeGPUNNPSDouble2DTestCase(DictBoxSortNNPS2DTestCase):
    """Test for Z-Order SFC based OpenCL algorithm"""

    def setUp(self):
        NNPS2DTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = True

        self.nps = gpu_nnps.OctreeGPUNNPS(
            dim=2, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )

    def tearDown(self):
        super(OctreeGPUNNPSDouble2DTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class NNPSTestCase(unittest.TestCase):
    """Standard nearest neighbor queries and comparison with the brute
    force approach.

    We randomly distribute particles in 3-space and compare the list
    of neighbors using the NNPS algorithms and the brute force
    approach.

    The following particle arrays are set up for testing
    1) pa1, pa2: uniformly distributed distribution with a constant h. pa1
    and p2 have a different number of particles and hence, a different h.
    2) pa3: Uniformly distributed distribution for both the coordinates and
    for h.
    3) pa4: h varies along with spatial coordinates.
    """

    def setUp(self):
        """Default set-up used by all the tests
        """
        numpy.random.seed(123)
        # Datasets with constant h
        self.numPoints1 = numPoints1 = 1 << 11
        self.numPoints2 = numPoints2 = 1 << 10

        # Datasets with varying h
        self.numPoints3 = numPoints3 = 1 << 10

        # FIXME: Tets fail with m4=9
        # Looks like the issue arises due to rounding errors which should be
        # acceptable to a degree. Need to modify tests or brute force NNPS to
        # handle such cases appropriately
        m4 = 8
        self.numPoints4 = numPoints4 = m4 ** 3

        self.pa1 = pa1 = self._create_random(numPoints1)
        self.pa2 = pa2 = self._create_random(numPoints2)
        self.pa3 = pa3 = self._create_random_variable_h(numPoints3)
        self.pa4 = pa4 = self._create_linear_radius(0.1, 0.4, m4)

        # the list of particles
        self.particles = [pa1, pa2, pa3, pa4]

    def _create_random(self, numPoints):
        # average particle spacing and volume in the unit cube
        dx = pow(1.0 / numPoints, 1. / 3.)

        # create random points in the interval [-1, 1]^3
        x1, y1, z1 = random.random((3, numPoints)) * 2.0 - 1.0
        h1 = numpy.ones_like(x1) * 1.2 * dx
        gid1 = numpy.arange(numPoints).astype(numpy.uint32)

        # first particle array
        pa = get_particle_array(
            x=x1, y=y1, z=z1, h=h1, gid=gid1)

        return pa

    def _create_linear_radius(self, dx_min, dx_max, m):
        n = m ** 3
        base = numpy.linspace(1., (dx_max / dx_min), m)
        hl = base * dx_min
        xl = numpy.cumsum(hl)
        x, y, z = numpy.meshgrid(xl, xl, xl)
        x, y, z = x.ravel(), y.ravel(), z.ravel()
        h1, h2, h3 = numpy.meshgrid(hl, hl, hl)
        h = (h1 ** 2 + h2 ** 2 + h3 ** 2) ** 0.5
        h = h.ravel()
        gid = numpy.arange(n).astype(numpy.uint32)

        pa = get_particle_array(
            x=x, y=y, z=z, h=h, gid=gid
        )

        return pa

    def _create_random_variable_h(self, num_points):
        # average particle spacing and volume in the unit cube
        dx = pow(1.0 / num_points, 1. / 3.)

        # create random points in the interval [-1, 1]^3
        x1, y1, z1 = random.random((3, num_points)) * 2.0 - 1.0
        h1 = numpy.ones_like(x1) * \
            numpy.random.uniform(1, 4, size=num_points) * 1.2 * dx
        gid1 = numpy.arange(num_points).astype(numpy.uint32)

        # first particle array
        pa = get_particle_array(
            x=x1, y=y1, z=z1, h=h1, gid=gid1)

        return pa

    def _assert_neighbors(self, nbrs_nnps, nbrs_brute_force):
        # ensure that the lengths of the arrays are the same
        if nbrs_nnps.length != nbrs_brute_force.length:
            print(nbrs_nnps.get_npy_array(), nbrs_brute_force.get_npy_array())
        self.assertEqual(nbrs_nnps.length, nbrs_brute_force.length)
        nnbrs = nbrs_nnps.length

        _nbrs1 = nbrs_nnps.get_npy_array()
        _nbrs2 = nbrs_brute_force.get_npy_array()

        # sort the neighbors
        nbrs1 = sorted(_nbrs1[:nnbrs])
        nbrs2 = _nbrs2
        nbrs2.sort()

        # check each neighbor
        for i in range(nnbrs):
            self.assertEqual(nbrs1[i], nbrs2[i])

    def _test_neighbors_by_particle(self, src_index, dst_index, dst_numPoints):
        # nnps and the two neighbor lists
        nps = self.nps
        nbrs1 = UIntArray()
        nbrs2 = UIntArray()

        nps.set_context(src_index, dst_index)

        # get the neighbors and sort the result
        for i in range(dst_numPoints):
            nps.get_nearest_particles(src_index, dst_index, i, nbrs1)
            nps.brute_force_neighbors(src_index, dst_index, i, nbrs2)

            # ensure that the neighbor lists are the same

            self._assert_neighbors(nbrs1, nbrs2)


class DictBoxSortNNPSTestCase(NNPSTestCase):
    """Test for the original box-sort algorithm"""

    def setUp(self):
        """
        Default setup and tests used for 3D NNPS tests

        We run the tests on the following pairs of particle arrays:

        Set 1) Same particle arrays. Both have constant h.
        1) a -> a
        2) b -> b

        Set 2) Different particle arrays with constant h.
        1) a -> b
        2) b -> a

        Set 3) Variable h
        1) c -> c
        2) d -> d

        We then repeat the above tests again to ensure that we get the
        correct results even when running NNPS repeatedly
        """
        NNPSTestCase.setUp(self)
        self.nps = nnps.DictBoxSortNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )

    def test_neighbors_aa(self):
        self._test_neighbors_by_particle(src_index=0, dst_index=0,
                                         dst_numPoints=self.numPoints1)

    def test_neighbors_ab(self):
        self._test_neighbors_by_particle(src_index=0, dst_index=1,
                                         dst_numPoints=self.numPoints2)

    def test_neighbors_ba(self):
        self._test_neighbors_by_particle(src_index=1, dst_index=0,
                                         dst_numPoints=self.numPoints1)

    def test_neighbors_bb(self):
        self._test_neighbors_by_particle(src_index=1, dst_index=1,
                                         dst_numPoints=self.numPoints2)

    def test_neighbors_cc(self):
        self._test_neighbors_by_particle(src_index=2, dst_index=2,
                                         dst_numPoints=self.numPoints3)

    def test_neighbors_dd(self):
        self._test_neighbors_by_particle(src_index=3, dst_index=3,
                                         dst_numPoints=self.numPoints4)

    def test_repeated(self):
        self.test_neighbors_aa()
        self.test_neighbors_ab()
        self.test_neighbors_ba()
        self.test_neighbors_bb()
        self.test_neighbors_cc()
        self.test_neighbors_dd()


class BoxSortNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for the original box-sort algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.BoxSortNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


class SpatialHashNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Spatial Hash algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.SpatialHashNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


class SingleLevelStratifiedHashNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Stratified hash algorithm with num_levels = 1"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.StratifiedHashNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


class MultipleLevelsStratifiedHashNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Stratified hash algorithm with num_levels = 2"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.StratifiedHashNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            num_levels=2
        )


class SingleLevelStratifiedSFCNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Stratified SFC algorithm with num_levels = 1"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.StratifiedSFCNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


class MultipleLevelsStratifiedSFCNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Stratified SFC algorithm with num_levels = 2"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.StratifiedSFCNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            num_levels=2
        )


class ExtendedSpatialHashNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Extended Spatial Hash algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.ExtendedSpatialHashNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


class OctreeNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Octree based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.OctreeNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class OctreeParallelNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Octree based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.OctreeNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            test_parallel=True
        )


class CellIndexingNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Cell Indexing based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.CellIndexingNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


class ZOrderNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Z-Order SFC based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.ZOrderNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


class ExtendedZOrderNNPSAsymmetricTestCase(DictBoxSortNNPSTestCase):
    """Test for asymmetric Extended Z-Order SFC based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.ExtendedZOrderNNPS(
            dim=3, particles=self.particles, radius_scale=2.0, H=1,
            asymmetric=True
        )


class ExtendedZOrderNNPSSymmetricTestCase(DictBoxSortNNPSTestCase):
    """Test for symmetric Extended Z-Order SFC based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.ExtendedZOrderNNPS(
            dim=3, particles=self.particles, radius_scale=2.0, H=1,
            asymmetric=False
        )


class ExtendedZOrderNNPSSubdividedAsymTestCase(DictBoxSortNNPSTestCase):
    """Test for asymmetric Extended Z-Order SFC based algorithm with
    subdivision"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.ExtendedZOrderNNPS(
            dim=3, particles=self.particles, radius_scale=2.0, H=3,
            asymmetric=True
        )

    @pytest.mark.xfail(reason="ExtendedZOrderNNPS failing for \
                       different dest and src index")
    def test_neighbors_ab(self):
        self._test_neighbors_by_particle(src_index=0, dst_index=1,
                                         dst_numPoints=self.numPoints2)

    @pytest.mark.xfail(reason="ExtendedZOrderNNPS failing for \
                       different dest and src index")
    def test_neighbors_ba(self):
        self._test_neighbors_by_particle(src_index=1, dst_index=0,
                                         dst_numPoints=self.numPoints1)

    @pytest.mark.xfail(reason="ExtendedZOrderNNPS failing for \
                       different dest and src index")
    def test_repeated(self):
        self.test_neighbors_aa()
        self.test_neighbors_ab()
        self.test_neighbors_ba()
        self.test_neighbors_bb()
        self.test_neighbors_cc()
        self.test_neighbors_dd()


class ExtendedZOrderNNPSSubdividedSymTestCase(
        ExtendedZOrderNNPSSubdividedAsymTestCase):
    """Test for symmetric Extended Z-Order SFC based algorithm with
    subdivision"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.ExtendedZOrderNNPS(
            dim=3, particles=self.particles, radius_scale=2.0, H=3,
            asymmetric=False
        )


class ZOrderGPUNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Z-Order SFC based OpenCL algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False

        self.nps = gpu_nnps.ZOrderGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )

    def tearDown(self):
        super(ZOrderGPUNNPSTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class ZOrderGPUNNPSTestCaseCUDA(ZOrderGPUNNPSTestCase):
    def setUp(self):
        NNPSTestCase.setUp(self)
        importorskip("pycuda")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False

        self.nps = gpu_nnps.ZOrderGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='cuda'
        )

    def tearDown(self):
        super(ZOrderGPUNNPSTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class BruteForceNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for OpenCL brute force algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False

        self.nps = gpu_nnps.BruteForceNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )

    def tearDown(self):
        super(BruteForceNNPSTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class OctreeGPUNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Z-Order SFC based OpenCL algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False

        self.nps = gpu_nnps.OctreeGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )

    def tearDown(self):
        super(OctreeGPUNNPSTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class ZOrderGPUDoubleNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Z-Order SFC based OpenCL algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = True
        self.nps = gpu_nnps.ZOrderGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )

    def tearDown(self):
        super(ZOrderGPUDoubleNNPSTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class ZOrderGPUDoubleNNPSTestCaseCUDA(ZOrderGPUDoubleNNPSTestCase):
    """Test for Z-Order SFC based OpenCL algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        importorskip("pycuda")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = True
        self.nps = gpu_nnps.ZOrderGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='cuda'
        )

    def tearDown(self):
        super(ZOrderGPUDoubleNNPSTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class OctreeGPUDoubleNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Octree based OpenCL algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = True
        self.nps = gpu_nnps.OctreeGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )

    def tearDown(self):
        super(OctreeGPUDoubleNNPSTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class TestZOrderGPUNNPSWithSorting(DictBoxSortNNPSTestCase):
    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False
        self.nps = gpu_nnps.ZOrderGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )
        self.nps.spatially_order_particles(0)
        self.nps.spatially_order_particles(1)

        for pa in self.particles:
            pa.gpu.pull()

    def tearDown(self):
        super(TestZOrderGPUNNPSWithSorting, self).tearDown()
        get_config().use_double = self._orig_use_double


class TestZOrderGPUNNPSWithSortingCUDA(TestZOrderGPUNNPSWithSorting):
    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pycuda")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False
        self.nps = gpu_nnps.ZOrderGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='cuda'
        )
        self.nps.spatially_order_particles(0)
        self.nps.spatially_order_particles(1)

        for pa in self.particles:
            pa.gpu.pull()

    def tearDown(self):
        super(TestZOrderGPUNNPSWithSorting, self).tearDown()
        get_config().use_double = self._orig_use_double


class OctreeGPUNNPSWithSortingTestCase(DictBoxSortNNPSTestCase):
    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False
        self.nps = gpu_nnps.OctreeGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            backend='opencl'
        )
        self.nps.spatially_order_particles(0)
        self.nps.spatially_order_particles(1)

        for pa in self.particles:
            pa.gpu.pull()

    def tearDown(self):
        super(OctreeGPUNNPSWithSortingTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class OctreeGPUNNPSWithPartitioningTestCase(DictBoxSortNNPSTestCase):
    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps
        cfg = get_config()
        self._orig_use_double = cfg.use_double
        cfg.use_double = False
        self.nps = gpu_nnps.OctreeGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            use_partitions=True, backend='opencl'
        )

        for pa in self.particles:
            pa.gpu.pull()

    def tearDown(self):
        super(OctreeGPUNNPSWithPartitioningTestCase, self).tearDown()
        get_config().use_double = self._orig_use_double


class StratifiedSFCGPUNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Stratified SFC based OpenCL algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        cl = importorskip("pyopencl")
        from pysph.base import gpu_nnps

        self.nps = gpu_nnps.StratifiedSFCGPUNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            num_levels=2, backend='opencl'
        )

    @pytest.mark.xfail(reason="StratifiedSFCGPUNNPS failing for \
                       variable h cases")
    def test_neighbors_dd(self):
        self._test_neighbors_by_particle(src_index=3, dst_index=3,
                                         dst_numPoints=self.numPoints4)

    @pytest.mark.xfail(reason="StratifiedSFCGPUNNPS failing for \
                       variable h cases")
    def test_repeated(self):
        self.test_neighbors_aa()
        self.test_neighbors_ab()
        self.test_neighbors_ba()
        self.test_neighbors_bb()
        self.test_neighbors_cc()
        self.test_neighbors_dd()


class CompressedOctreeNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Compressed Octree based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.CompressedOctreeNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
class CompressedOctreeParallelNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for Parallel Compressed Octree based algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.CompressedOctreeNNPS(
            dim=3, particles=self.particles, radius_scale=2.0,
            test_parallel=True
        )


class LinkedListNNPSTestCase(DictBoxSortNNPSTestCase):
    """Test for the original box-sort algorithm"""

    def setUp(self):
        NNPSTestCase.setUp(self)
        self.nps = nnps.LinkedListNNPS(
            dim=3, particles=self.particles, radius_scale=2.0
        )

    def test_cell_index_positivity(self):
        nps = self.nps
        ncells_tot = nps.ncells_tot
        ncells_per_dim = nps.ncells_per_dim
        dim = nps.dim

        # cell indices should be positive. We iterate over the
        # flattened indices, get the unflattened version and check
        # that each component remains positive
        for cell_index in range(ncells_tot):
            cid = nnps.py_unflatten(cell_index, ncells_per_dim, dim)

            self.assertTrue(cid.x > -1)
            self.assertTrue(cid.y > -1)
            self.assertTrue(cid.z > -1)


class TestNNPSOnLargeDomain(unittest.TestCase):
    def _make_particles(self, nx=20):
        x, y, z = numpy.random.random((3, nx, nx, nx))
        x = numpy.ravel(x)
        y = numpy.ravel(y)
        z = numpy.ravel(z)
        h = numpy.ones_like(x) * 1.3 / nx

        pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=h)
        # Place one particle far far away
        # On Linux and OSX this works even if sz is 100000.
        # However, on Windows this fails but works with 1000,
        # hence we set it to 1000.
        sz = 1000.0
        pa.add_particles(x=[sz], y=[sz], z=[sz])
        return pa

    def test_linked_list_nnps_raises_exception_for_large_domain(self):
        # Given/When
        pa = self._make_particles(20)
        # Then
        self.assertRaises(
            RuntimeError, nnps.LinkedListNNPS, dim=3, particles=[pa],
            cache=True
        )

    def test_box_sort_works_for_large_domain(self):
        # Given
        pa = self._make_particles(20)
        # We turn on cache so it computes all the neighbors quickly for us.
        nps = nnps.BoxSortNNPS(dim=3, particles=[pa], cache=True)
        nbrs = UIntArray()
        direct = UIntArray()
        nps.set_context(0, 0)
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nps.brute_force_neighbors(0, 0, i, direct)
            x = nbrs.get_npy_array()
            y = direct.get_npy_array()
            x.sort()
            y.sort()
            assert numpy.all(x == y)

    def test_spatial_hash_works_for_large_domain(self):
        # Given
        pa = self._make_particles(20)
        # We turn on cache so it computes all the neighbors quickly for us.
        nps = nnps.SpatialHashNNPS(dim=3, particles=[pa], cache=True)
        nbrs = UIntArray()
        direct = UIntArray()
        nps.set_context(0, 0)
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nps.brute_force_neighbors(0, 0, i, direct)
            x = nbrs.get_npy_array()
            y = direct.get_npy_array()
            x.sort()
            y.sort()
            assert numpy.all(x == y)

    def test_extended_spatial_hash_works_for_large_domain(self):
        # Given
        pa = self._make_particles(20)
        # We turn on cache so it computes all the neighbors quickly for us.
        nps = nnps.ExtendedSpatialHashNNPS(dim=3, particles=[pa], cache=True)
        nbrs = UIntArray()
        direct = UIntArray()
        nps.set_context(0, 0)
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nps.brute_force_neighbors(0, 0, i, direct)
            x = nbrs.get_npy_array()
            y = direct.get_npy_array()
            x.sort()
            y.sort()
            assert numpy.all(x == y)

    def test_octree_works_for_large_domain(self):
        # Given
        pa = self._make_particles(20)
        # We turn on cache so it computes all the neighbors quickly for us.
        nps = nnps.OctreeNNPS(dim=3, particles=[pa], cache=True)
        nbrs = UIntArray()
        direct = UIntArray()
        nps.set_context(0, 0)
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nps.brute_force_neighbors(0, 0, i, direct)
            x = nbrs.get_npy_array()
            y = direct.get_npy_array()
            x.sort()
            y.sort()
            assert numpy.all(x == y)

    @mark.skipif(get_number_of_threads() == 1,
                 reason="N_threads=1; OpenMP does not seem available.")
    def test_octree_parallel_works_for_large_domain(self):
        # Given
        pa = self._make_particles(20)
        # We turn on cache so it computes all the neighbors quickly for us.
        nps = nnps.OctreeNNPS(dim=3, particles=[pa], cache=True,
                              test_parallel=True)
        nbrs = UIntArray()
        direct = UIntArray()
        nps.set_context(0, 0)
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nps.brute_force_neighbors(0, 0, i, direct)
            x = nbrs.get_npy_array()
            y = direct.get_npy_array()
            x.sort()
            y.sort()
            assert numpy.all(x == y)

    def test_compressed_octree_works_for_large_domain(self):
        # Given
        pa = self._make_particles(20)
        # We turn on cache so it computes all the neighbors quickly for us.
        nps = nnps.CompressedOctreeNNPS(dim=3, particles=[pa], cache=True)
        nbrs = UIntArray()
        direct = UIntArray()
        nps.set_context(0, 0)
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nps.brute_force_neighbors(0, 0, i, direct)
            x = nbrs.get_npy_array()
            y = direct.get_npy_array()
            x.sort()
            y.sort()
            assert numpy.all(x == y)

    @mark.skipif(get_number_of_threads() == 1,
                 reason="N_threads=1; OpenMP does not seem available.")
    def test_compressed_octree_parallel_works_for_large_domain(self):
        # Given
        pa = self._make_particles(20)
        # We turn on cache so it computes all the neighbors quickly for us.
        nps = nnps.CompressedOctreeNNPS(dim=3, particles=[pa], cache=True,
                                        test_parallel=True)
        nbrs = UIntArray()
        direct = UIntArray()
        nps.set_context(0, 0)
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nps.brute_force_neighbors(0, 0, i, direct)
            x = nbrs.get_npy_array()
            y = direct.get_npy_array()
            x.sort()
            y.sort()
            assert numpy.all(x == y)


class TestLinkedListNNPSWithSorting(unittest.TestCase):
    def _make_particles(self, nx=20):
        x = numpy.linspace(0, 1, nx)
        h = numpy.ones_like(x) / (nx - 1)

        pa = get_particle_array(name='fluid', x=x, h=h)
        nps = nnps.LinkedListNNPS(dim=1, particles=[pa], sort_gids=True)
        return pa, nps

    def test_nnps_sorts_without_gids(self):
        # Given
        pa, nps = self._make_particles(10)

        # When
        nps.set_context(0, 0)
        # Test the that gids are actually huge and invalid.
        self.assertEqual(numpy.max(pa.gid), numpy.min(pa.gid))
        self.assertTrue(numpy.max(pa.gid) > pa.gid.size)

        # Then
        nbrs = UIntArray()
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nb = nbrs.get_npy_array()
            sorted_nbrs = sorted(nb.copy())
            self.assertTrue(numpy.all(nb == sorted_nbrs))

    def test_nnps_sorts_with_valid_gids(self):
        # Given
        pa, nps = self._make_particles(10)
        pa.gid[:] = numpy.arange(pa.x.size)
        nps.update()

        # When
        nps.set_context(0, 0)
        # Test the that gids are actually valid.
        self.assertEqual(numpy.max(pa.gid), pa.gid.size - 1)
        self.assertEqual(numpy.min(pa.gid), 0)

        # Then
        nbrs = UIntArray()
        for i in range(pa.get_number_of_particles()):
            nps.get_nearest_particles(0, 0, i, nbrs)
            nb = nbrs.get_npy_array()
            sorted_nbrs = sorted(nb.copy())
            self.assertTrue(numpy.all(nb == sorted_nbrs))


class TestSpatialHashNNPSWithSorting(TestLinkedListNNPSWithSorting):
    def _make_particles(self, nx=20):
        x = numpy.linspace(0, 1, nx)
        h = numpy.ones_like(x) / (nx - 1)

        pa = get_particle_array(name='fluid', x=x, h=h)
        nps = nnps.SpatialHashNNPS(dim=1, particles=[pa], sort_gids=True)
        return pa, nps


class TestMultipleLevelsStratifiedHashNNPSWithSorting(
        TestLinkedListNNPSWithSorting):
    def _make_particles(self, nx=20):
        x = numpy.linspace(0, 1, nx)
        h = numpy.ones_like(x) / (nx - 1)

        pa = get_particle_array(name='fluid', x=x, h=h)
        nps = nnps.StratifiedHashNNPS(dim=1, particles=[pa], num_levels=2,
                                      sort_gids=True)
        return pa, nps


class TestMultipleLevelsStratifiedSFCNNPSWithSorting(
        TestLinkedListNNPSWithSorting):
    def _make_particles(self, nx=20):
        x = numpy.linspace(0, 1, nx)
        h = numpy.ones_like(x) / (nx - 1)

        pa = get_particle_array(name='fluid', x=x, h=h)
        nps = nnps.StratifiedSFCNNPS(dim=1, particles=[pa], num_levels=2,
                                     sort_gids=True)
        return pa, nps


def test_large_number_of_neighbors_linked_list():
    x = numpy.random.random(1 << 14) * 0.1
    y = x.copy()
    z = x.copy()
    h = numpy.ones_like(x)
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=h)

    nps = nnps.LinkedListNNPS(dim=3, particles=[pa], cache=False)
    nbrs = UIntArray()
    nps.get_nearest_particles(0, 0, 0, nbrs)
    # print(nbrs.length)
    assert nbrs.length == len(x)


def test_neighbor_cache_update_doesnt_leak():
    # Given
    x, y, z = numpy.random.random((3, 1000))
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=0.05)

    nps = nnps.LinkedListNNPS(dim=3, particles=[pa], cache=True)
    nps.set_context(0, 0)
    nps.cache[0].find_all_neighbors()
    old_length = sum(x.length for x in nps.cache[0]._neighbor_arrays)

    # When
    nps.update()
    nps.set_context(0, 0)
    nps.cache[0].find_all_neighbors()

    # Then
    new_length = sum(x.length for x in nps.cache[0]._neighbor_arrays)
    assert new_length == old_length


nnps_classes = [
    nnps.BoxSortNNPS,
    nnps.CellIndexingNNPS,
    nnps.CompressedOctreeNNPS,
    nnps.ExtendedSpatialHashNNPS,
    nnps.LinkedListNNPS,
    nnps.OctreeNNPS,
    nnps.SpatialHashNNPS,
    nnps.StratifiedHashNNPS,
    nnps.StratifiedSFCNNPS,
    nnps.ZOrderNNPS
]


@pytest.mark.parametrize("cls", nnps_classes)
def test_corner_case_1d_few_cells(cls):
    x, y, z = [0.131, 0.359], [1.544, 1.809], [-3.6489999, -2.8559999]
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=1.0)
    nbrs = UIntArray()
    bf_nbrs = UIntArray()
    nps = cls(dim=3, particles=[pa], radius_scale=0.7)
    for i in range(2):
        nps.get_nearest_particles(0, 0, i, nbrs)
        nps.brute_force_neighbors(0, 0, i, bf_nbrs)
        assert sorted(nbrs) == sorted(bf_nbrs), 'Failed for particle: %d' % i


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
def test_corner_case_parallel_octree_1d_few_cells():
    x, y, z = [0.131, 0.359], [1.544, 1.809], [-3.6489999, -2.8559999]
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=1.0)
    nbrs = UIntArray()
    bf_nbrs = UIntArray()
    nps = nnps.OctreeNNPS(dim=3, particles=[pa], radius_scale=0.7,
                          test_parallel=True)
    for i in range(2):
        nps.get_nearest_particles(0, 0, i, nbrs)
        nps.brute_force_neighbors(0, 0, i, bf_nbrs)
        assert sorted(nbrs) == sorted(bf_nbrs), 'Failed for particle: %d' % i


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
def test_corner_case_parallel_compressed_octree_1d_few_cells():
    x, y, z = [0.131, 0.359], [1.544, 1.809], [-3.6489999, -2.8559999]
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=1.0)
    nbrs = UIntArray()
    bf_nbrs = UIntArray()
    nps = nnps.CompressedOctreeNNPS(dim=3, particles=[pa], radius_scale=0.7,
                                    test_parallel=True)
    for i in range(2):
        nps.get_nearest_particles(0, 0, i, nbrs)
        nps.brute_force_neighbors(0, 0, i, bf_nbrs)
        assert sorted(nbrs) == sorted(bf_nbrs), 'Failed for particle: %d' % i


def test_use_2d_for_1d_data_with_llnps():
    y = numpy.array([1.0, 1.5])
    h = numpy.ones_like(y)
    pa = get_particle_array(name='fluid', y=y, h=h)
    nps = nnps.LinkedListNNPS(dim=2, particles=[pa], cache=False)
    nbrs = UIntArray()
    nps.get_nearest_particles(0, 0, 0, nbrs)
    print(nbrs.length)
    assert nbrs.length == len(y)


def test_use_3d_for_1d_data_with_llnps():
    y = numpy.array([1.0, 1.5])
    h = numpy.ones_like(y)
    pa = get_particle_array(name='fluid', y=y, h=h)
    nps = nnps.LinkedListNNPS(dim=3, particles=[pa], cache=False)
    nbrs = UIntArray()
    nps.get_nearest_particles(0, 0, 0, nbrs)
    print(nbrs.length)
    assert nbrs.length == len(y)


def test_large_number_of_neighbors_spatial_hash():
    x = numpy.random.random(1 << 14) * 0.1
    y = x.copy()
    z = x.copy()
    h = numpy.ones_like(x)
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=h)

    nps = nnps.SpatialHashNNPS(dim=3, particles=[pa], cache=False)
    nbrs = UIntArray()
    nps.get_nearest_particles(0, 0, 0, nbrs)
    # print(nbrs.length)
    assert nbrs.length == len(x)


def test_large_number_of_neighbors_octree():
    x = numpy.random.random(1 << 14) * 0.1
    y = x.copy()
    z = x.copy()
    h = numpy.ones_like(x)
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=h)

    nps = nnps.OctreeNNPS(dim=3, particles=[pa], cache=False)
    nbrs = UIntArray()
    nps.get_nearest_particles(0, 0, 0, nbrs)
    # print(nbrs.length)
    assert nbrs.length == len(x)


@mark.skipif(get_number_of_threads() == 1,
             reason="N_threads=1; OpenMP does not seem available.")
def test_large_number_of_neighbors_parallel_octree():
    x = numpy.random.random(1 << 14) * 0.1
    y = x.copy()
    z = x.copy()
    h = numpy.ones_like(x)
    pa = get_particle_array(name='fluid', x=x, y=y, z=z, h=h)

    nps = nnps.OctreeNNPS(dim=3, particles=[pa], cache=False,
                          test_parallel=True)
    nbrs = UIntArray()
    nps.get_nearest_particles(0, 0, 0, nbrs)
    # print(nbrs.length)
    assert nbrs.length == len(x)


def test_flatten_unflatten():
    # first consider the 2D case where we assume a 4 X 5 grid of cells
    dim = 2
    ncells_per_dim = IntArray(3)
    ncells_per_dim[0] = 4
    ncells_per_dim[1] = 5
    ncells_per_dim[2] = 0

    # valid un-flattened cell indices
    cids = [[i, j] for i in range(4) for j in range(5)]
    for _cid in cids:
        cid = IntPoint(_cid[0], _cid[1], 0)
        flattened = nnps.py_flatten(cid, ncells_per_dim, dim)
        unflattened = nnps.py_unflatten(flattened, ncells_per_dim, dim)

        # the unflattened index should match with cid
        assert (cid == unflattened)

    # 3D
    dim = 3
    ncells_per_dim = IntArray(3)
    ncells_per_dim[0] = 4
    ncells_per_dim[1] = 5
    ncells_per_dim[2] = 2

    # valid un-flattened indices
    cids = [[i, j, k] for i in range(4) for j in range(5) for k in range(2)]
    for _cid in cids:
        cid = IntPoint(_cid[0], _cid[1], _cid[2])
        flattened = nnps.py_flatten(cid, ncells_per_dim, dim)
        unflattened = nnps.py_unflatten(flattened, ncells_per_dim, dim)

        # the unflattened index should match with cid
        assert (cid == unflattened)


def test_1D_get_valid_cell_index():
    dim = 1

    # simulate a dummy distribution such that 10 cells are along the
    # 'x' direction
    n_cells = 10
    ncells_per_dim = IntArray(3)

    ncells_per_dim[0] = n_cells
    ncells_per_dim[1] = 1
    ncells_per_dim[2] = 1

    # target cell
    cx = 1
    cy = cz = 0

    # as long as cy and cz are 0, the function should return the valid
    # flattened cell index for the cell
    for i in [-1, 0, 1]:
        index = nnps.py_get_valid_cell_index(
            IntPoint(cx + i, cy, cz), ncells_per_dim, dim, n_cells)
        assert index != -1

    # index should be -1 whenever cy and cz are > 1. This is
    # specifically the case that was failing earlier.
    for j in [-1, 1]:
        for k in [-1, 1]:
            index = nnps.py_get_valid_cell_index(
                IntPoint(cx, cy + j, cz + k), ncells_per_dim, dim, n_cells)
            assert index == -1

    # When the cx > n_cells or < -1 it should be invalid
    for i in [-2, -1, n_cells, n_cells + 1]:
        index = nnps.py_get_valid_cell_index(
            IntPoint(i, cy, cz), ncells_per_dim, dim, n_cells)
        assert index == -1


def test_get_centroid():
    cell = nnps.Cell(IntPoint(0, 0, 0), cell_size=0.1, narrays=1)
    centroid = Point()
    cell.get_centroid(centroid)

    assert (abs(centroid.x - 0.05) < 1e-10)
    assert (abs(centroid.y - 0.05) < 1e-10)
    assert (abs(centroid.z - 0.05) < 1e-10)

    cell = nnps.Cell(IntPoint(1, 2, 3), cell_size=0.5, narrays=1)
    cell.get_centroid(centroid)

    assert (abs(centroid.x - 0.75) < 1e-10)
    assert (abs(centroid.y - 1.25) < 1e-10)
    assert (abs(centroid.z - 1.75) < 1e-10)


def test_get_bbox():
    cell_size = 0.1
    cell = nnps.Cell(IntPoint(0, 0, 0), cell_size=cell_size, narrays=1)
    centroid = Point()
    boxmin = Point()
    boxmax = Point()

    cell.get_centroid(centroid)
    cell.get_bounding_box(boxmin, boxmax)

    assert (abs(boxmin.x - (centroid.x - 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmin.y - (centroid.y - 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmin.z - (centroid.z - 1.5 * cell_size)) < 1e-10)

    assert (abs(boxmax.x - (centroid.x + 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmax.y - (centroid.y + 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmax.z - (centroid.z + 1.5 * cell_size)) < 1e-10)

    cell_size = 0.5
    cell = nnps.Cell(IntPoint(1, 2, 0), cell_size=cell_size, narrays=1)

    cell.get_centroid(centroid)
    cell.get_bounding_box(boxmin, boxmax)

    assert (abs(boxmin.x - (centroid.x - 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmin.y - (centroid.y - 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmin.z - (centroid.z - 1.5 * cell_size)) < 1e-10)

    assert (abs(boxmax.x - (centroid.x + 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmax.y - (centroid.y + 1.5 * cell_size)) < 1e-10)
    assert (abs(boxmax.z - (centroid.z + 1.5 * cell_size)) < 1e-10)


if __name__ == '__main__':
    unittest.main()



import pytest
from pytest import approx

import pysph.sph.gas_dynamics.riemann_solver as R

solvers = [
    R.ducowicz, R.exact, R.hll_ball, R.hllc,
    R.hllc_ball, R.hlle, R.hllsy, R.llxf, R.roe,
    R.van_leer
]


def _check_shock_tube(solver, **approx_kw):
    # Given
    # Shock tube
    gamma = 1.4
    rhol, pl, ul = 1.0, 1.0, 0.0
    rhor, pr, ur = 0.125, 0.1, 0.0

    result = [0.0, 0.0]
    # When
    solver(rhol, rhor, pl, pr, ul, ur, gamma, niter=20,
           tol=1e-6, result=result)

    # Then
    assert result == approx((0.30313, 0.92745), **approx_kw)


def _check_blastwave(solver, **approx_kw):
    # Given
    gamma = 1.4
    result = [0.0, 0.0]
    rhol, pl, ul = 1.0, 1000.0, 0.0
    rhor, pr, ur = 1.0, 0.01, 0.0

    # When
    solver(rhol, rhor, pl, pr, ul, ur, gamma, niter=20,
           tol=1e-6, result=result)

    # Then
    assert result == approx((460.894, 19.5975), **approx_kw)


def _check_sjogreen(solver, **approx_kw):
    # Given
    gamma = 1.4
    result = [0.0, 0.0]
    rhol, pl, ul = 1.0, 0.4, -2.0
    rhor, pr, ur = 1.0, 0.4, 2.0

    # When
    solver(rhol, rhor, pl, pr, ul, ur, gamma, niter=20,
           tol=1e-6, result=result)

    # Then
    assert result == approx((0.0018938, 0.0), **approx_kw)


def _check_woodward_collela(solver, **approx_kw):
    # Given
    gamma = 1.4
    result = [0.0, 0.0]
    rhol, pl, ul = 1.0, 0.01, 0.0
    rhor, pr, ur = 1.0, 100.0, 0.0

    # When
    solver(rhol, rhor, pl, pr, ul, ur, gamma, niter=20,
           tol=1e-6, result=result)

    # Then
    assert result == approx((46.0950, -6.19633), **approx_kw)


def test_exact_riemann():
    solver = R.exact

    _check_shock_tube(solver, rel=1e-4)
    _check_blastwave(solver, rel=1e-3)
    _check_sjogreen(solver, abs=1e-4)
    _check_woodward_collela(solver, rel=1e-4)


def test_van_leer():
    solver = R.van_leer
    _check_shock_tube(solver, rel=1e-3)
    _check_blastwave(solver, rel=1e-2)
    _check_sjogreen(solver, abs=1e-2)
    _check_woodward_collela(solver, rel=1e-2)


def test_ducowicz():
    solver = R.ducowicz
    _check_shock_tube(solver, rel=0.2)
    _check_blastwave(solver, rel=0.4)
    _check_sjogreen(solver, abs=1e-2)
    _check_woodward_collela(solver, rel=0.4)


# Most other solvers seem rather poor in comparison.
@pytest.mark.parametrize("solver", solvers)
def test_all_solver_api(solver):
    if solver.__name__ in ['roe', 'hllc']:
        rel = 2.0
    else:
        rel = 1.0
    _check_shock_tube(solver, rel=rel)



"""Rectangular dam break over a dry bed with particle splitting and merging
(19 min).

The case without particle splitting is described in "A corrected smooth
particle hydrodynamics formulation of the shallow-water equations", Miguel
Rodriguez-Paz and Javier Bonet, Computers & Structures, Vol 83, pp 1396-1410
(2005). DOI:10.1016/j.compstruc.2004.11.025

"""
import os

# Numpy
from numpy import (ones_like, mgrid, array, sort, linspace, arange,
                   concatenate, savez)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        GatherDensityEvalNextIteration, NonDimensionalDensityResidual,
        UpdateSmoothingLength, CheckConvergenceDensityResidual, SWEOS,
        SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration, ParticleSplit, CheckForParticlesToSplit,
        DaughterVelocityEval, FindMergeable, InitialDensityEvalAfterMerge
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator


# Constants
rho_w = 1000.0
g = 9.81
dim = 2


class RectangularDamBreakSplitandMerge(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.5,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=0.025,
            help="Spacing between the particles")
        group.add_argument(
            "--dw0", action="store", type=float, dest="dw0", default=1.0,
            help="Initial depth of the fluid column (m)")
        group.add_argument(
            "--le", action="store", type=float, dest="le", default=2.0,
            help="Initial length of the fluid column (m)")
        group.add_argument(
            "--w", action="store", type=float, dest="w", default=1.0,
            help="Initial width of the fluid column (m)")
        group.add_argument(
            "--coeff_A_split", action="store", type=float,
            dest="coeff_A_max", default=1.0, help="Ratio of area of\
            particle to initial area after which the particle splitting is\
            activated")
        group.add_argument(
            "--coeff_h_split", action="store", type=float, dest="coeff_h_max",
            default=2.0, help="Ratio of smoothing length of particle to\
            initial smoothing length after which the particle splitting is\
            deactivated")
        group.add_argument(
            "--coeff_A_merge", action="store", type=float,
            dest="coeff_A_min", default=1.0, help="Ratio of area of\
            particle to initial area below which the particle merging is\
            activated")
        group.add_argument(
            "--x_min_split", action="store", type=float,
            dest="x_min_split", default=2.0, help="Position after which\
            splitting is activated")
        group.add_argument(
            "--x_max_split", action="store", type=float,
            dest="x_max_split", default=3.0, help="Position after which\
            splitting is deactivated")
        group.add_argument(
            "--x_min_merge", action="store", type=float,
            dest="x_min_merge", default=3.0, help="Position after which\
            merging is activated")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.dw0 = self.options.dw0
        self.le = self.options.le
        self.w = self.options.w

        self.coeff_A_max = self.options.coeff_A_max
        self.coeff_A_min = self.options.coeff_A_min
        self.coeff_h_max = self.options.coeff_h_max

        # Particle splitting activated when (A_i > A_max and h_i < h_max) and
        # (x_min_split < x_i < x_max_split)
        self.A_max = self.coeff_A_max * self.dx**2
        self.h_max = self.coeff_h_max * self.hdx * self.dx
        self.x_min_split = self.options.x_min_split
        self.x_max_split = self.options.x_max_split

        # Particle merging activated when A_i < A_min and x_i > x_min_merge
        self.A_min = self.coeff_A_min * self.dx**2
        self.x_min_merge = self.options.x_min_merge

    def create_particles(self):
        """Create the Rectangular patch of fluid."""
        # The wall at x = 0 is simulated by applying a symmetry condition
        # in the fluid, i.e., an additional column of fluid to the left of the
        # wall
        x, y = mgrid[-self.le:self.le+1e-4:self.dx,
                     -(self.w)/2.:(self.w)/2.+1e-4:self.dx]
        x = x.ravel()
        y = y.ravel()

        dx = self.dx
        d = self.dw0
        hdx = self.hdx

        m = ones_like(x) * dx * dx * rho_w * d
        h = ones_like(x) * hdx * dx
        h0 = ones_like(x) * hdx * dx

        rho = ones_like(x) * rho_w * d
        rho0 = ones_like(x) * rho_w * d

        pa = gpa_swe(x=x, y=y, m=m, rho0=rho0, rho=rho, h=h, h0=h0,
                     name='fluid')

        compute_initial_props([pa])
        return [pa]

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = SWEIntegrator(fluid=SWEStep())
        tf = 1.0
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            cfl=0.3,
            adaptive_timestep=True,
            output_at_times=[0.1, 0.2, 0.4, 0.6],
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    Group(
                        equations=[
                            GatherDensityEvalNextIteration(dest='fluid',
                                                           sources=['fluid'])
                            ]
                        ),
                    Group(
                        equations=[
                            NonDimensionalDensityResidual(dest='fluid')
                            ]
                        ),
                    Group(
                        equations=[
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                            ], update_nnps=True
                        ),
                    Group(
                        equations=[
                            CheckConvergenceDensityResidual(dest='fluid')
                            ],
                    )], iterate=True, max_iterations=10
            ),
            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(dest='fluid',
                                                            sources=['fluid'])
                    ]
                ),
            Group(
                equations=[
                    DaughterVelocityEval(rhow=rho_w, dest='fluid',
                                         sources=['fluid'])
                    ]
                ),
            Group(
                equations=[
                    SWEOS(dest='fluid')
                    ]
                ),
            Group(
                equations=[
                    ParticleAcceleration(dim=dim, dest='fluid',
                                         sources=['fluid'], u_only=True)
                    ]
                ),
            Group(
                equations=[
                   FindMergeable(dest='fluid', sources=['fluid'],
                                 A_min=self.A_min, x_min=self.x_min_merge)
                   ], update_nnps=True
                 ),
            Group(
                equations=[
                        InitialDensityEvalAfterMerge(dest='fluid',
                                                     sources=['fluid'])
                    ]
                 ),
            Group(
                equations=[
                    Group(
                        equations=[
                            GatherDensityEvalNextIteration(dest='fluid',
                                                           sources=['fluid'])
                            ]
                        ),
                    Group(
                        equations=[
                            NonDimensionalDensityResidual(dest='fluid')
                            ]
                        ),
                    Group(
                        equations=[
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                            ], update_nnps=True
                        ),
                    Group(
                        equations=[
                            CheckConvergenceDensityResidual(dest='fluid')
                            ],
                    )], iterate=True, max_iterations=100
            ),
            Group(
                equations=[
                    CheckForParticlesToSplit(
                        dest='fluid', h_max=self.h_max, A_max=self.A_max,
                        x_min=self.x_min_split, x_max=self.x_max_split)
                    ],
                ),
            ]
        return equations

    def pre_step(self, solver):
        for pa in self.particles:
            ps = ParticleSplit(pa)
            ps.do_particle_split()
        self.nnps.update()

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_depth_of_dam()

    def _plot_depth_of_dam(self):
        from matplotlib import pyplot as plt
        t_to_plot = [0.1, 0.2, 0.4, 0.6]
        fname_for_plot = []
        for fname in self.output_files:
            data = load(fname)
            t = data['solver_data']['t']
            if t in t_to_plot:
                fname_for_plot.append(fname)

        def pos_actual(dw0, dw, t):
            # Given the initial depth (dw0) and depth (dw) of a fluid column
            # section at time t, returns the exact location of that section
            # with respect to the gate position (Ritters Solution)
            # Note: This solution is valid till the expansion wave reaches the
            # wall
            g = 9.8
            return (2*(g*dw0)**0.5 - 3*(g*dw)**0.5)*t

        num = 0
        for fname in fname_for_plot:
            data = load(fname)
            fluid = data['arrays']['fluid']

            y = arange(-(self.w)/2., (self.w)/2.+1e-4, self.dx)
            y_srt = sort(y)
            len_y = len(y_srt)
            idx_mid_y = int(len_y/2.)
            mid_y = y_srt[idx_mid_y]

            # Mid section of fluid column along positive x axis
            cond1 = abs(fluid.y - mid_y) < 1e-10
            cond2 = fluid.x >= 0
            cond = cond1 & cond2

            x = fluid.x[cond]
            dw = fluid.dw[cond]
            u = fluid.u[cond]
            rho = fluid.rho[cond]
            au = fluid.au[cond]

            if len(x) == 0:
                return

            x_ind_sort = x.argsort()
            x = x[x_ind_sort]
            x_max = max(x)
            dw = dw[x_ind_sort]
            u = u[x_ind_sort]
            rho = rho[x_ind_sort]
            au = au[x_ind_sort]

            # To get the remaining analytical curve of pos vs depth
            dw_rem = linspace(dw[-1], 0, 50)
            dw_actual = concatenate((dw, dw_rem), axis=0)

            x_actual = pos_actual(self.dw0, dw_actual, t_to_plot[num])
            x_actual = concatenate((array([0]), x_actual+self.le), axis=0)

            dw_actual = concatenate(
                    (array([self.options.dw0]), dw_actual), axis=0)

            fname_res = os.path.join(
                    self.output_dir, 'results%d.npz' % (num+1))
            savez(
                fname_res, t=t_to_plot[num], x=x, x_actual=x_actual, y=y,
                dw=dw, dw_actual=dw_actual, u=u, rho=rho, au=au)

            plt.clf()
            plt.plot(x_actual, dw_actual, 'r-', label='Analytical Solution')
            plt.plot(x, dw, 'bo ', label='SWE-SPH', markersize=2)
            plt.ylim(0, 1.2*self.dw0)
            plt.xlim(0, 1.5*x_max)
            plt.legend()
            plt.xlabel('x (m)')
            plt.ylabel('dw (m)')
            plt.title('t = %0.1fs' % t_to_plot[num])
            fig = os.path.join(self.output_dir, "depth%d.png" % (num+1))
            plt.savefig(fig, dpi=300)
            num += 1


def compute_initial_props(particles):
    one_time_equations = [
                Group(
                    equations=[
                        SWEOS(dest='fluid')
                        ]
                    ),
            ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == '__main__':
    app = RectangularDamBreakSplitandMerge()
    app.run()
    app.post_process(app.info_filename)



"""Incompressible flow past a periodic lattice of cylinders. (30 minutes)
"""

import os
# numpy
import numpy as np

# PySPH imports
from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application

from pysph.sph.scheme import TVFScheme


# domain and reference values
L = 0.1
Umax = 5e-5
c0 = 10 * Umax
rho0 = 1000.0
p0 = c0 * c0 * rho0
a = 0.02
H = L
fx = 1.5e-7

# Reynolds number and kinematic viscosity
Re = 1.0
nu = a * Umax / Re

# Numerical setup
nx = 100
dx = L / nx
ghost_extent = 5 * 1.5 * dx
hdx = 1.0

# adaptive time steps
h0 = hdx * dx
dt_cfl = 0.25 * h0 / (c0 + Umax)
dt_viscous = 0.125 * h0**2 / nu
dt_force = 0.25 * np.sqrt(h0 / abs(fx))

tf = 1000.0
dt = min(dt_cfl, dt_viscous, dt_force)


class LatticeCylinders(Application):
    def create_domain(self):
        # domain for periodicity
        domain = DomainManager(
            xmin=0, xmax=L, ymin=0, ymax=H, periodic_in_x=True,
            periodic_in_y=True
        )
        return domain

    def create_particles(self):
        # create all the particles
        _x = np.arange(dx / 2, L, dx)
        _y = np.arange(dx / 2, H, dx)
        x, y = np.meshgrid(_x, _y)
        x = x.ravel()
        y = y.ravel()

        # sort out the fluid and the solid
        indices = []
        cx = 0.5 * L
        cy = 0.5 * H
        for i in range(x.size):
            xi = x[i]
            yi = y[i]
            if (np.sqrt((xi - cx)**2 + (yi - cy)**2) > a):
                    # if ( (yi > 0) and (yi < H) ):
                indices.append(i)

        # create the arrays
        solid = get_particle_array(name='solid', x=x, y=y)

        # remove the fluid particles from the solid
        fluid = solid.extract_particles(indices)
        fluid.set_name('fluid')
        solid.remove_particles(indices)

        print("Periodic cylinders :: Re = %g, nfluid = %d, nsolid=%d, dt = %g" % (
            Re, fluid.get_number_of_particles(),
            solid.get_number_of_particles(), dt))

        self.scheme.setup_properties([fluid, solid])

        # setup the particle properties
        volume = dx * dx

        # mass is set to get the reference density of rho0
        fluid.m[:] = volume * rho0
        solid.m[:] = volume * rho0
        solid.rho[:] = rho0

        # reference pressures and densities
        fluid.rho[:] = rho0

        # volume is set as dx^2
        fluid.V[:] = 1. / volume
        solid.V[:] = 1. / volume

        # smoothing lengths
        fluid.h[:] = hdx * dx
        solid.h[:] = hdx * dx

        # return the particle list
        return [fluid, solid]

    def create_scheme(self):
        s = TVFScheme(
            ['fluid'], ['solid'], dim=2, rho0=rho0, c0=c0, nu=nu,
            p0=p0, pb=p0, h0=dx * hdx, gx=fx
        )
        s.configure_solver(tf=tf, dt=dt)
        return s

    def post_process(self, info_fname):
        info = self.read_info(info_fname)
        if len(self.output_files) == 0 or self.rank > 0:
            return

        y, ui_lby2, ui_l, xx, yy, vmag = self._plot_velocity()
        res = os.path.join(self.output_dir, "results.npz")
        np.savez(res, y=y, ui_l=ui_l, ui_lby2=ui_lby2, xx=xx, yy=yy, vmag=vmag)

    def _plot_velocity(self):
        from pysph.tools.interpolator import Interpolator
        from pysph.solver.utils import load

        # Find the u profile for comparison.
        y = np.linspace(0.0, H, 100)
        x = np.ones_like(y) * L / 2
        fname = self.output_files[-1]
        data = load(fname)
        dm = self.create_domain()
        interp = Interpolator(list(data['arrays'].values()), x=x, y=y,
                              domain_manager=dm)
        ui_lby2 = interp.interpolate('u')
        x = np.ones_like(y) * L
        interp.set_interpolation_points(x=x, y=y)
        ui_l = interp.interpolate('u')

        import matplotlib
        matplotlib.use('Agg')

        from matplotlib import pyplot as plt
        y /= H
        y -= 0.5
        f = plt.figure()
        plt.plot(y, ui_lby2, 'k-', label='x=L/2')
        plt.plot(y, ui_l, 'k-', label='x=L')
        plt.xlabel('y/H')
        plt.ylabel('u')
        plt.legend()
        fig = os.path.join(self.output_dir, 'u_profile.png')
        plt.savefig(fig, dpi=300)
        plt.close()

        # Plot the contours of vmag.
        xx, yy = np.mgrid[0:L:100j, 0:H:100j]
        interp.set_interpolation_points(x=xx, y=yy)
        u = interp.interpolate('u')
        v = interp.interpolate('v')
        xx /= L
        yy /= H
        vmag = np.sqrt(u * u + v * v)
        f = plt.figure()
        plt.contourf(xx, yy, vmag)
        plt.xlabel('x/L')
        plt.ylabel('y/H')
        plt.colorbar()
        fig = os.path.join(self.output_dir, 'vmag_contour.png')
        plt.savefig(fig, dpi=300)
        plt.close()

        return y, ui_lby2, ui_l, xx, yy, vmag


if __name__ == '__main__':
    app = LatticeCylinders()
    app.run()
    app.post_process(app.info_filename)



"""Four cubes bouncing inside a box. (10 seconds)

This is used to test the rigid body equations and the support for multiple
bodies.
"""

import numpy as np

from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_rigid_body
from pysph.sph.equation import Group

from pysph.sph.integrator import EPECIntegrator

from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.sph.rigid_body import (BodyForce, RigidBodyCollision,
                                  RigidBodyMoments, RigidBodyMotion,
                                  RK2StepRigidBody)

dim = 3

dt = 5e-3
tf = 5.0
gz = -9.81

hdx = 1.0
rho0 = 100.0


def make_cube(lx, ly, lz, dx):
    """Return points x, y, z for a cube centered at origin with given lengths.
    """
    # Convert to floats to be safe with integer division.
    lx, ly, lz = list(map(float, (lx, ly, lz)))
    x, y, z = np.mgrid[-lx / 2:lx / 2 + dx:dx, -ly / 2:ly / 2 + dx:dx, -lz / 2:
                       lz / 2 + dx:dx]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    return x, y, z


class BouncingCubes(Application):
    def create_particles(self):
        dx = 1.0 / 9.0
        _x, _y, _z = make_cube(0.5, 0.5, 0.5, dx)
        _z += 1.0
        _id = np.ones(_x.shape, dtype=int)
        x, y, z, body_id = [], [], [], []
        disp = [(0.4, 0, 0), (-0.4, 0, 0), (0.0, 1.0, 0.0), (0.0, -1.0, 0.0)]
        for i, d in enumerate(disp):
            x.append(_x + d[0])
            y.append(_y + d[1])
            z.append(_z + d[2])
            body_id.append(_id * i)
        x = np.concatenate(x)
        y = np.concatenate(y)
        z = np.concatenate(z)
        body_id = np.concatenate(body_id)
        m = np.ones_like(x) * dx * dx * dx * rho0
        h = np.ones_like(x) * hdx * dx

        # Split this one cube

        # radius of each sphere constituting in cube
        rad_s = np.ones_like(x) * dx
        body = get_particle_array_rigid_body(
            name='body', x=x, y=y, z=z, h=h, rho=rho0,
            m=m, body_id=body_id, rad_s=rad_s
        )

        body.vc[0] = 5.0
        body.vc[2] = -5.0
        body.vc[6] = -5.0
        body.vc[7] = -5.0
        body.vc[10] = 5.0

        # Create the tank.
        nx, ny, nz = 40, 40, 40
        xmin, xmax, ymin, ymax, zmin, zmax = -2.1, 2.1, -2.1, 2.1, -2.1, 2.1
        dx = (xmax - xmin) / (nx - 1)
        x, y, z = np.mgrid[xmin:xmax:nx * 1j, ymin:ymax:ny * 1j, zmin:zmax:nz *
                           1j]
        interior = ((x < 1.8) & (x > -1.8)) & ((y < 1.8) & (y > -1.8)) & (
            (z > -1.8) & (z <= 2))
        tank = np.logical_not(interior)
        x = x[tank].flat
        y = y[tank].flat
        z = z[tank].flat
        m = np.ones_like(x) * dx * dx * dx * rho0
        h = np.ones_like(x) * hdx * dx

        # radius of each sphere constituting in cube
        rad_s = np.ones_like(x) * dx
        tank = get_particle_array_rigid_body(name='tank', x=x, y=y, z=z, h=h,
                                             m=m, rad_s=rad_s, rho=rho0)
        tank.total_mass[0] = np.sum(m)

        return [body, tank]

    def create_solver(self):
        kernel = CubicSpline(dim=dim)

        integrator = EPECIntegrator(body=RK2StepRigidBody())

        solver = Solver(kernel=kernel, dim=dim, integrator=integrator, dt=dt,
                        tf=tf, adaptive_timestep=False)
        solver.set_print_freq(10)
        return solver

    def create_equations(self):
        equations = [
            Group(equations=[
                BodyForce(dest='body', sources=None, gz=gz),
                RigidBodyCollision(dest='body', sources=['tank', 'body'],
                                   kn=1e4, en=0.8)
            ]),
            Group(equations=[RigidBodyMoments(dest='body', sources=None)]),
            Group(equations=[RigidBodyMotion(dest='body', sources=None)]),
        ]
        return equations


if __name__ == '__main__':
    app = BouncingCubes()
    app.run()
